import yt

# first add density contours on a density slice
ds = yt.load("GasSloshing/sloshing_nomag2_hdf5_plt_cnt_0150")

# add density contours on the density slice.
p = yt.SlicePlot(ds, "x", ("gas", "density"))
p.annotate_contour(("gas", "density"))
p.save()

# then add temperature contours on the same density slice
p = yt.SlicePlot(ds, "x", ("gas", "density"))
p.annotate_contour(("gas", "temperature"))
p.save(str(ds) + "_T_contour")



import yt
import yt.units as u

ds = yt.load("HiresIsolatedGalaxy/DD0044/DD0044")

center = [0.53, 0.53, 0.53]
normal = [0, 0, 1]
radius = 40 * u.kpc
height = 2 * u.kpc

disk = ds.disk(center, [0, 0, 1], radius, height)

profile = yt.create_profile(
    data_source=disk,
    bin_fields=[("index", "radius"), ("gas", "velocity_cylindrical_theta")],
    fields=[("gas", "mass")],
    n_bins=256,
    units=dict(radius="kpc", velocity_cylindrical_theta="km/s", mass="Msun"),
    logs=dict(radius=False, velocity_cylindrical_theta=False),
    weight_field=None,
    extrema=dict(radius=(0, 40), velocity_cylindrical_theta=(-250, 250)),
)

plot = yt.PhasePlot.from_profile(profile)
plot.set_cmap(("gas", "mass"), "YlOrRd")

plot.save()



from collections import defaultdict

import numpy as np

from yt.funcs import is_sequence
from yt.geometry.grid_container import GridTree, MatchPointsToGrids
from yt.utilities.exceptions import (
    YTInconsistentGridFieldShape,
    YTInconsistentGridFieldShapeGridDims,
    YTInconsistentParticleFieldShape,
)
from yt.utilities.logger import ytLogger as mylog

from .fields import StreamFieldInfo


def assign_particle_data(ds, pdata, bbox):

    """
    Assign particle data to the grids using MatchPointsToGrids. This
    will overwrite any existing particle data, so be careful!
    """

    particle_index_fields = [
        f"particle_position_{ax}" for ax in ds.coordinates.axis_order
    ]
    for ptype in ds.particle_types_raw:
        check_fields = [(ptype, pi_field) for pi_field in particle_index_fields]
        check_fields.append((ptype, "particle_position"))
        if all(f not in pdata for f in check_fields):
            pdata_ftype = {}
            for f in [k for k in sorted(pdata)]:
                if not hasattr(pdata[f], "shape"):
                    continue
                if f == "number_of_particles":
                    continue
                mylog.debug("Reassigning '%s' to ('%s','%s')", f, ptype, f)
                pdata_ftype[ptype, f] = pdata.pop(f)
            pdata_ftype.update(pdata)
            pdata = pdata_ftype

    # Note: what we need to do here is a bit tricky.  Because occasionally this
    # gets called before we property handle the field detection, we cannot use
    # any information about the index.  Fortunately for us, we can generate
    # most of the GridTree utilizing information we already have from the
    # stream handler.

    if len(ds.stream_handler.fields) > 1:
        pdata.pop("number_of_particles", None)
        num_grids = len(ds.stream_handler.fields)
        parent_ids = ds.stream_handler.parent_ids
        num_children = np.zeros(num_grids, dtype="int64")
        # We're going to do this the slow way
        mask = np.empty(num_grids, dtype="bool")
        for i in range(num_grids):
            np.equal(parent_ids, i, mask)
            num_children[i] = mask.sum()
        levels = ds.stream_handler.levels.astype("int64").ravel()
        grid_tree = GridTree(
            num_grids,
            ds.stream_handler.left_edges,
            ds.stream_handler.right_edges,
            ds.stream_handler.dimensions,
            ds.stream_handler.parent_ids,
            levels,
            num_children,
        )

        grid_pdata = []
        for _ in range(num_grids):
            grid = {"number_of_particles": 0}
            grid_pdata.append(grid)
            particle_index_fields = [
                f"particle_position_{ax}" for ax in ds.coordinates.axis_order
            ]

        for ptype in ds.particle_types_raw:
            if (ptype, "particle_position_x") in pdata:
                # we call them x, y, z even though they may be different field names
                x, y, z = (pdata[ptype, pi_field] for pi_field in particle_index_fields)
            elif (ptype, "particle_position") in pdata:
                x, y, z = pdata[ptype, "particle_position"].T
            else:
                raise KeyError(
                    "Cannot decompose particle data without position fields!"
                )
            pts = MatchPointsToGrids(grid_tree, len(x), x, y, z)
            particle_grid_inds = pts.find_points_in_tree()
            (assigned_particles,) = (particle_grid_inds >= 0).nonzero()
            num_particles = particle_grid_inds.size
            num_unassigned = num_particles - assigned_particles.size
            if num_unassigned > 0:
                eps = np.finfo(x.dtype).eps
                s = np.array(
                    [
                        [x.min() - eps, x.max() + eps],
                        [y.min() - eps, y.max() + eps],
                        [z.min() - eps, z.max() + eps],
                    ]
                )
                sug_bbox = [
                    [min(bbox[0, 0], s[0, 0]), max(bbox[0, 1], s[0, 1])],
                    [min(bbox[1, 0], s[1, 0]), max(bbox[1, 1], s[1, 1])],
                    [min(bbox[2, 0], s[2, 0]), max(bbox[2, 1], s[2, 1])],
                ]
                mylog.warning(
                    "Discarding %s particles (out of %s) that are outside "
                    "bounding box. Set bbox=%s to avoid this in the future.",
                    num_unassigned,
                    num_particles,
                    sug_bbox,
                )
                particle_grid_inds = particle_grid_inds[assigned_particles]
                x = x[assigned_particles]
                y = y[assigned_particles]
                z = z[assigned_particles]
            idxs = np.argsort(particle_grid_inds)
            particle_grid_count = np.bincount(
                particle_grid_inds.astype("intp"), minlength=num_grids
            )
            particle_indices = np.zeros(num_grids + 1, dtype="int64")
            if num_grids > 1:
                np.add.accumulate(
                    particle_grid_count.squeeze(), out=particle_indices[1:]
                )
            else:
                particle_indices[1] = particle_grid_count.squeeze()
            for i, pcount in enumerate(particle_grid_count):
                grid_pdata[i]["number_of_particles"] += pcount
                start = particle_indices[i]
                end = particle_indices[i + 1]
                for key in pdata.keys():
                    if key[0] == ptype:
                        grid_pdata[i][key] = pdata[key][idxs][start:end]

    else:
        grid_pdata = [pdata]

    for pd, gi in zip(grid_pdata, sorted(ds.stream_handler.fields)):
        ds.stream_handler.fields[gi].update(pd)
        ds.stream_handler.particle_types.update(set_particle_types(pd))
        npart = ds.stream_handler.fields[gi].pop("number_of_particles", 0)
        ds.stream_handler.particle_count[gi] = npart


def process_data(data, grid_dims=None, allow_callables=True):
    new_data, field_units = {}, {}
    for field, val in data.items():
        # val is a data array
        if isinstance(val, np.ndarray):
            # val is a YTArray
            if hasattr(val, "units"):
                field_units[field] = val.units
                new_data[field] = val.copy().d
            # val is a numpy array
            else:
                field_units[field] = ""
                new_data[field] = val.copy()

        # val is a tuple of (data, units)
        elif isinstance(val, tuple) and len(val) == 2:
            try:
                valid_data = isinstance(val[0], np.ndarray)
                if allow_callables:
                    valid_data = valid_data or callable(val[0])
                assert isinstance(field, (str, tuple)), "Field name is not a string!"
                assert (
                    valid_data
                ), "Field data is not an ndarray or callable (with nproc == 1)!"
                assert isinstance(val[1], str), "Unit specification is not a string!"
                field_units[field] = val[1]
                new_data[field] = val[0]
            except AssertionError as e:
                raise RuntimeError("The data dict appears to be invalid.\n" + str(e))

        # val is a list of data to be turned into an array
        elif is_sequence(val):
            field_units[field] = ""
            new_data[field] = np.asarray(val)

        elif callable(val):
            if not allow_callables:
                raise RuntimeError(
                    "Callable functions can not be specified "
                    "in conjunction with nprocs > 1."
                )
            field_units[field] = ""
            new_data[field] = val
        else:
            raise RuntimeError(
                "The data dict appears to be invalid. "
                "The data dictionary must map from field "
                "names to (numpy array, unit spec) tuples. "
            )

    data = new_data

    # At this point, we have arrays for all our fields
    new_data = {}
    for field in data:
        n_shape = 3
        if not callable(data[field]):
            n_shape = len(data[field].shape)
        if isinstance(field, tuple):
            new_field = field
        elif n_shape in (1, 2):
            new_field = ("io", field)
        elif n_shape == 3:
            new_field = ("stream", field)
        else:
            raise RuntimeError
        new_data[new_field] = data[field]
        field_units[new_field] = field_units.pop(field)
        known_fields = (
            StreamFieldInfo.known_particle_fields + StreamFieldInfo.known_other_fields
        )
        # We do not want to override any of the known ones, if it's not
        # overridden here.
        if (
            any(f[0] == new_field[1] for f in known_fields)
            and field_units[new_field] == ""
        ):
            field_units.pop(new_field)
    data = new_data
    # Sanity checking that all fields have the same dimensions.
    g_shapes = []
    p_shapes = defaultdict(list)
    for field in data:
        if callable(data[field]):
            continue
        f_shape = data[field].shape
        n_shape = len(f_shape)
        if n_shape in (1, 2):
            p_shapes[field[0]].append((field[1], f_shape[0]))
        elif n_shape == 3:
            g_shapes.append((field, f_shape))
    if len(g_shapes) > 0:
        g_s = np.array([s[1] for s in g_shapes])
        if not np.all(g_s == g_s[0]):
            raise YTInconsistentGridFieldShape(g_shapes)
        if grid_dims is not None:
            if not np.all(g_s == grid_dims):
                raise YTInconsistentGridFieldShapeGridDims(g_shapes, grid_dims)
    if len(p_shapes) > 0:
        for ptype, p_shape in p_shapes.items():
            p_s = np.array([s[1] for s in p_shape])
            if not np.all(p_s == p_s[0]):
                raise YTInconsistentParticleFieldShape(ptype, p_shape)
    # Now that we know the particle fields are consistent, determine the number
    # of particles.
    if len(p_shapes) > 0:
        number_of_particles = np.sum([s[0][1] for s in p_shapes.values()])
    else:
        number_of_particles = 0
    return field_units, data, number_of_particles


def set_particle_types(data):
    particle_types = {}
    for key in data.keys():
        if key == "number_of_particles":
            continue
        elif callable(data[key]):
            particle_types[key] = False
        elif len(data[key].shape) == 1:
            particle_types[key] = True
        else:
            particle_types[key] = False
    return particle_types



import yt

# Load the dataset.
ds = yt.load("GalaxyClusterMerger/fiducial_1to3_b0.273d_hdf5_plt_cnt_0175")

# Create a data object that represents the whole box.
ad = ds.all_data()

# This is identical to the simple phase plot, except we supply
# the fractional=True keyword to divide the profile data by the sum.
plot = yt.PhasePlot(
    ad,
    ("gas", "density"),
    ("gas", "temperature"),
    ("gas", "mass"),
    weight_field=None,
    fractional=True,
)

# Save the image.
# Optionally, give a string as an argument
# to name files with a keyword.
plot.save()



import fileinput
import io
import os
import warnings
import zipfile
from functools import partial, wraps
from re import finditer
from tempfile import NamedTemporaryFile, TemporaryFile
from typing import Tuple

import numpy as np
from more_itertools import always_iterable
from tqdm import tqdm

from yt._maintenance.deprecation import issue_deprecation_warning
from yt.config import ytcfg
from yt.data_objects.field_data import YTFieldData
from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer1D,
    YTSelectionContainer2D,
    YTSelectionContainer3D,
)
from yt.fields.field_exceptions import NeedsGridType, NeedsOriginalGrid
from yt.frontends.sph.data_structures import ParticleDataset
from yt.funcs import (
    get_memory_usage,
    is_sequence,
    iter_fields,
    mylog,
    only_on_root,
    validate_moment,
)
from yt.geometry import particle_deposit as particle_deposit
from yt.geometry.coordinates.cartesian_coordinates import all_data
from yt.loaders import load_uniform_grid
from yt.units._numpy_wrapper_functions import uconcatenate
from yt.units.unit_object import Unit  # type: ignore
from yt.units.yt_array import YTArray
from yt.utilities.exceptions import (
    YTNoAPIKey,
    YTNotInsideNotebook,
    YTParticleDepositionNotImplemented,
    YTTooManyVertices,
)
from yt.utilities.grid_data_format.writer import write_to_gdf
from yt.utilities.lib.cyoctree import CyOctree
from yt.utilities.lib.interpolators import ghost_zone_interpolate
from yt.utilities.lib.marching_cubes import march_cubes_grid, march_cubes_grid_flux
from yt.utilities.lib.misc_utilities import fill_region, fill_region_float
from yt.utilities.lib.pixelization_routines import (
    interpolate_sph_grid_gather,
    interpolate_sph_positions_gather,
    normalization_1d_utility,
    normalization_3d_utility,
    pixelize_sph_kernel_arbitrary_grid,
)
from yt.utilities.lib.quad_tree import QuadTree
from yt.utilities.math_utils import compute_stddev_image
from yt.utilities.minimal_representation import MinimalProjectionData
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    communication_system,
    parallel_objects,
    parallel_root_only,
)
from yt.visualization.color_maps import get_colormap_lut


class YTStreamline(YTSelectionContainer1D):
    """
    This is a streamline, which is a set of points defined as
    being parallel to some vector field.

    This object is typically accessed through the Streamlines.path
    function.  The resulting arrays have their dimensionality
    reduced to one, and an ordered list of points at an (x,y)
    tuple along `axis` are available, as is the `t` field, which
    corresponds to a unitless measurement along the ray from start
    to end.

    Parameters
    ----------
    positions : array-like
        List of streamline positions
    length : float
        The magnitude of the distance; dts will be divided by this
    fields : list of strings, optional
        If you want the object to pre-retrieve a set of fields, supply them
        here.  This is not necessary.
    ds : dataset object
        Passed in to access the index
    kwargs : dict of items
        Any additional values are passed as field parameters that can be
        accessed by generated fields.

    Examples
    --------

    >>> from yt.visualization.api import Streamlines
    >>> streamlines = Streamlines(ds, [0.5] * 3)
    >>> streamlines.integrate_through_volume()
    >>> stream = streamlines.path(0)
    >>> fig, ax = plt.subplots()
    >>> ax.set_yscale("log")
    >>> ax.plot(stream["t"], stream[("gas", "density")], "-x")
    """

    _type_name = "streamline"
    _con_args = ("positions",)
    sort_by = "t"

    def __init__(self, positions, length=1.0, fields=None, ds=None, **kwargs):
        YTSelectionContainer1D.__init__(self, ds, fields, **kwargs)
        self.positions = positions
        self.dts = np.empty_like(positions[:, 0])
        self.dts[:-1] = np.sqrt(
            np.sum((self.positions[1:] - self.positions[:-1]) ** 2, axis=1)
        )
        self.dts[-1] = self.dts[-1]
        self.length = length
        self.dts /= length
        self.ts = np.add.accumulate(self.dts)
        self._set_center(self.positions[0])
        self.set_field_parameter("center", self.positions[0])
        self._dts, self._ts = {}, {}
        # self._refresh_data()

    def _get_list_of_grids(self):
        # Get the value of the line at each LeftEdge and RightEdge
        LE = self.ds.grid_left_edge
        RE = self.ds.grid_right_edge
        # Check left faces first
        min_streampoint = np.min(self.positions, axis=0)
        max_streampoint = np.max(self.positions, axis=0)
        p = np.all((min_streampoint <= RE) & (max_streampoint > LE), axis=1)
        self._grids = self.index.grids[p]

    def _get_data_from_grid(self, grid, field):
        # No child masking here; it happens inside the mask cut
        mask = self._get_cut_mask(grid)
        if field == "dts":
            return self._dts[grid.id]
        if field == "t":
            return self._ts[grid.id]
        return grid[field].flat[mask]

    def _get_cut_mask(self, grid):
        points_in_grid = np.all(self.positions > grid.LeftEdge, axis=1) & np.all(
            self.positions <= grid.RightEdge, axis=1
        )
        pids = np.where(points_in_grid)[0]
        mask = np.zeros(points_in_grid.sum(), dtype="int64")
        dts = np.zeros(points_in_grid.sum(), dtype="float64")
        ts = np.zeros(points_in_grid.sum(), dtype="float64")
        for mi, (i, pos) in enumerate(zip(pids, self.positions[points_in_grid])):
            if not points_in_grid[i]:
                continue
            ci = ((pos - grid.LeftEdge) / grid.dds).astype("int")
            if grid.child_mask[ci[0], ci[1], ci[2]] == 0:
                continue
            for j in range(3):
                ci[j] = min(ci[j], grid.ActiveDimensions[j] - 1)
            mask[mi] = np.ravel_multi_index(ci, grid.ActiveDimensions)
            dts[mi] = self.dts[i]
            ts[mi] = self.ts[i]
        self._dts[grid.id] = dts
        self._ts[grid.id] = ts
        return mask


class YTProj(YTSelectionContainer2D):
    _key_fields = YTSelectionContainer2D._key_fields + ["weight_field"]
    _con_args = ("axis", "field", "weight_field")
    _container_fields = ("px", "py", "pdx", "pdy", "weight_field")

    def __init__(
        self,
        field,
        axis,
        weight_field=None,
        center=None,
        ds=None,
        data_source=None,
        method="integrate",
        field_parameters=None,
        max_level=None,
        *,
        moment=1,
        style=None,
    ):
        super().__init__(axis, ds, field_parameters)
        if style is not None:
            issue_deprecation_warning(
                "The 'style' keyword argument is a deprecated alias for 'method'. "
                "Please use method directly.",
                since="3.2",
                removal="4.2",
                stacklevel=4,
            )
            method = style
        if method == "mip":
            issue_deprecation_warning(
                "The 'mip' method value is a deprecated alias for 'max'. "
                "Please use max directly.",
                since="4.1",
                stacklevel=4,
            )
            method = "max"
        if method == "sum":
            self.method = "integrate"
            self._sum_only = True
        else:
            self.method = method
            self._sum_only = False
        if self.method in ["max", "mip"]:
            self.func = np.max
        elif self.method == "min":
            self.func = np.min
        elif self.method == "integrate":
            self.func = np.sum  # for the future
        else:
            raise NotImplementedError(self.method)
        validate_moment(moment, weight_field)
        self.moment = moment
        self._set_center(center)
        self._projected_units = {}
        if data_source is None:
            data_source = self.ds.all_data()
        if max_level is not None:
            data_source.max_level = max_level
        for k, v in data_source.field_parameters.items():
            if k not in self.field_parameters or self._is_default_field_parameter(k):
                self.set_field_parameter(k, v)
        self.data_source = data_source
        if weight_field is None:
            self.weight_field = weight_field
        else:
            self.weight_field = self._determine_fields(weight_field)[0]

        for f in self._determine_fields(field):
            nodal_flag = self.ds._get_field_info(f).nodal_flag
            if any(nodal_flag):
                raise RuntimeError(
                    "Nodal fields are currently not supported for projections."
                )

    @property
    def blocks(self):
        return self.data_source.blocks

    @property
    def field(self):
        return [k for k in self.field_data.keys() if k not in self._container_fields]

    def get_data(self, fields=None):
        fields = self._determine_fields(fields)
        sfields = []
        if self.moment == 2:

            def _sq_field(field, data, fname: Tuple[str, str]):
                return data[fname] ** 2

            for fname in fields:
                fd = self.ds._get_field_info(*fname)
                self.ds.add_field(
                    (fname[0], f"tmp_{fname[1]}_squared"),
                    partial(_sq_field, fname=fname),
                    sampling_type=fd.sampling_type,
                    units=f"({fd.units})*({fd.units})",
                )
                sfields.append((fname[0], f"tmp_{fname[1]}_squared"))
        nfields = len(fields)
        nsfields = len(sfields)
        # We need a new tree for every single set of fields we add
        if nfields == 0:
            return
        if isinstance(self.ds, ParticleDataset):
            return
        tree = self._get_tree(nfields + nsfields)
        # This only needs to be done if we are in parallel; otherwise, we can
        # safely build the mesh as we go.
        if communication_system.communicators[-1].size > 1:
            for chunk in self.data_source.chunks([], "io", local_only=False):
                self._initialize_chunk(chunk, tree)
        _units_initialized = False
        with self.data_source._field_parameter_state(self.field_parameters):
            for chunk in parallel_objects(
                self.data_source.chunks([], "io", local_only=True)
            ):
                if not _units_initialized:
                    self._initialize_projected_units(fields, chunk)
                    _units_initialized = True
                self._handle_chunk(chunk, fields + sfields, tree)
        # if there's less than nprocs chunks, units won't be initialized
        # on all processors, so sync with _projected_units on rank 0
        projected_units = self.comm.mpi_bcast(self._projected_units)
        self._projected_units = projected_units
        # Note that this will briefly double RAM usage
        if self.method in ["max", "mip"]:
            merge_style = -1
            op = "max"
        elif self.method == "min":
            merge_style = -2
            op = "min"
        elif self.method == "integrate":
            merge_style = 1
            op = "sum"
        else:
            raise NotImplementedError
        # TODO: Add the combine operation
        xax = self.ds.coordinates.x_axis[self.axis]
        yax = self.ds.coordinates.y_axis[self.axis]

        ix, iy, ires, nvals, nwvals = tree.get_all(False, merge_style)
        px, pdx = self.ds.index._icoords_to_fcoords(
            ix[:, None], ires // self.ds.ires_factor, axes=(xax,)
        )
        py, pdy = self.ds.index._icoords_to_fcoords(
            iy[:, None], ires // self.ds.ires_factor, axes=(yax,)
        )
        px = px.ravel()
        py = py.ravel()
        pdx = pdx.ravel()
        pdy = pdy.ravel()
        np.multiply(pdx, 0.5, pdx)
        np.multiply(pdy, 0.5, pdy)

        nvals = self.comm.mpi_allreduce(nvals, op=op)
        nwvals = self.comm.mpi_allreduce(nwvals, op=op)
        if self.weight_field is not None:
            # If there are 0s remaining in the weight vals
            # this will not throw an error, but silently
            # return nans for vals where dividing by 0
            # Leave as NaNs to be auto-masked by Matplotlib
            with np.errstate(invalid="ignore"):
                np.divide(nvals, nwvals[:, None], nvals)
        # We now convert to half-widths and center-points
        data = {}
        code_length = self.ds.domain_width.units
        data["px"] = self.ds.arr(px, code_length)
        data["py"] = self.ds.arr(py, code_length)
        data["weight_field"] = nwvals
        data["pdx"] = self.ds.arr(pdx, code_length)
        data["pdy"] = self.ds.arr(pdy, code_length)
        data["fields"] = nvals
        # Now we run the finalizer, which is ignored if we don't need it
        field_data = np.hsplit(data.pop("fields"), nfields + nsfields)
        for fi, field in enumerate(fields):
            mylog.debug("Setting field %s", field)
            input_units = self._projected_units[field]
            fvals = field_data[fi].ravel()
            if self.moment == 2:
                fvals = compute_stddev_image(field_data[fi + nfields].ravel(), fvals)
            self[field] = self.ds.arr(fvals, input_units)
        for i in list(data.keys()):
            self[i] = data.pop(i)
        mylog.info("Projection completed")
        if self.moment == 2:
            for field in sfields:
                self.ds.field_info.pop(field)
        self.tree = tree

    def to_pw(self, fields=None, center="c", width=None, origin="center-window"):
        r"""Create a :class:`~yt.visualization.plot_window.PWViewerMPL` from this
        object.

        This is a bare-bones mechanism of creating a plot window from this
        object, which can then be moved around, zoomed, and on and on.  All
        behavior of the plot window is relegated to that routine.
        """
        pw = self._get_pw(fields, center, width, origin, "Projection")
        return pw

    def plot(self, fields=None):
        if hasattr(self.data_source, "left_edge") and hasattr(
            self.data_source, "right_edge"
        ):
            left_edge = self.data_source.left_edge
            right_edge = self.data_source.right_edge
            center = (left_edge + right_edge) / 2.0
            width = right_edge - left_edge
            xax = self.ds.coordinates.x_axis[self.axis]
            yax = self.ds.coordinates.y_axis[self.axis]
            lx, rx = left_edge[xax], right_edge[xax]
            ly, ry = left_edge[yax], right_edge[yax]
            width = (rx - lx), (ry - ly)
        else:
            width = self.ds.domain_width
            center = self.ds.domain_center
        pw = self._get_pw(fields, center, width, "native", "Projection")
        try:
            pw.show()
        except YTNotInsideNotebook:
            pass
        return pw

    def _initialize_projected_units(self, fields, chunk):
        for field in self.data_source._determine_fields(fields):
            if field in self._projected_units:
                continue
            finfo = self.ds._get_field_info(*field)
            if finfo.units is None:
                # First time calling a units="auto" field, infer units and cache
                # for future field accesses.
                finfo.units = str(chunk[field].units)
            field_unit = Unit(finfo.output_units, registry=self.ds.unit_registry)
            if self.method in ("min", "max") or self._sum_only:
                path_length_unit = Unit(registry=self.ds.unit_registry)
            else:
                ax_name = self.ds.coordinates.axis_name[self.axis]
                path_element_name = ("index", f"path_element_{ax_name}")
                path_length_unit = self.ds.field_info[path_element_name].units
                path_length_unit = Unit(
                    path_length_unit, registry=self.ds.unit_registry
                )
                # Only convert to appropriate unit system for path
                # elements that aren't angles
                if not path_length_unit.is_dimensionless:
                    path_length_unit = path_length_unit.get_base_equivalent(
                        unit_system=self.ds.unit_system
                    )
            if self.weight_field is None:
                self._projected_units[field] = field_unit * path_length_unit
            else:
                self._projected_units[field] = field_unit


class YTParticleProj(YTProj):
    """
    A projection operation optimized for SPH particle data.
    """

    _type_name = "particle_proj"

    def __init__(
        self,
        field,
        axis,
        weight_field=None,
        center=None,
        ds=None,
        data_source=None,
        method="integrate",
        field_parameters=None,
        max_level=None,
        *,
        moment=1,
        style=None,
    ):
        super().__init__(
            field,
            axis,
            weight_field,
            center,
            ds,
            data_source,
            method,
            field_parameters,
            max_level,
            moment=moment,
            style=style,
        )

    def _handle_chunk(self, chunk, fields, tree):
        raise NotImplementedError("Particle projections have not yet been implemented")


class YTQuadTreeProj(YTProj):
    """
    This is a data object corresponding to a line integral through the
    simulation domain.

    This object is typically accessed through the `proj` object that
    hangs off of index objects.  YTQuadTreeProj is a projection of a
    `field` along an `axis`.  The field can have an associated
    `weight_field`, in which case the values are multiplied by a weight
    before being summed, and then divided by the sum of that weight; the
    two fundamental modes of operating are direct line integral (no
    weighting) and average along a line of sight (weighting.)  What makes
    `proj` different from the standard projection mechanism is that it
    utilizes a quadtree data structure, rather than the old mechanism for
    projections.  It will not run in parallel, but serial runs should be
    substantially faster.  Note also that lines of sight are integrated at
    every projected finest-level cell.

    Parameters
    ----------
    field : string
        This is the field which will be "projected" along the axis.  If
        multiple are specified (in a list) they will all be projected in
        the first pass.
    axis : int
        The axis along which to slice.  Can be 0, 1, or 2 for x, y, z.
    weight_field : string
        If supplied, the field being projected will be multiplied by this
        weight value before being integrated, and at the conclusion of the
        projection the resultant values will be divided by the projected
        `weight_field`.
    center : array_like, optional
        The 'center' supplied to fields that use it.  Note that this does
        not have to have `coord` as one value.  Strictly optional.
    data_source : `yt.data_objects.data_containers.YTSelectionContainer`, optional
        If specified, this will be the data source used for selecting
        regions to project.
    method : string, optional
        The method of projection to be performed.
        "integrate" : integration along the axis
        "mip" : maximum intensity projection (deprecated)
        "max" : maximum intensity projection
        "min" : minimum intensity projection
        "sum" : same as "integrate", except that we don't multiply by the path length
        WARNING: The "sum" option should only be used for uniform resolution grid
        datasets, as other datasets may result in unphysical images.
    style : string, optional
        The same as the method keyword.  Deprecated as of version 3.0.2.
        Please use method keyword instead.
    field_parameters : dict of items
        Values to be passed as field parameters that can be
        accessed by generated fields.
    moment : integer, optional
        for a weighted projection, moment = 1 (the default) corresponds to a
        weighted average. moment = 2 corresponds to a weighted standard
        deviation.

    Examples
    --------

    >>> ds = load("RedshiftOutput0005")
    >>> prj = ds.proj(("gas", "density"), 0)
    >>> print(proj[("gas", "density")])
    """

    _type_name = "quad_proj"

    def __init__(
        self,
        field,
        axis,
        weight_field=None,
        center=None,
        ds=None,
        data_source=None,
        method="integrate",
        field_parameters=None,
        max_level=None,
        *,
        moment=1,
        style=None,
    ):
        super().__init__(
            field,
            axis,
            weight_field,
            center,
            ds,
            data_source,
            method,
            field_parameters,
            max_level,
            moment=moment,
            style=style,
        )

        if not self.deserialize(field):
            self.get_data(field)
            self.serialize()

    @property
    def _mrep(self):
        return MinimalProjectionData(self)

    def deserialize(self, fields):
        if not ytcfg.get("yt", "serialize"):
            return False
        for field in fields:
            self[field] = None
        deserialized_successfully = False
        store_file = self.ds.parameter_filename + ".yt"
        if os.path.isfile(store_file):
            deserialized_successfully = self._mrep.restore(store_file, self.ds)

            if deserialized_successfully:
                mylog.info("Using previous projection data from %s", store_file)
                for field, field_data in self._mrep.field_data.items():
                    self[field] = field_data
        if not deserialized_successfully:
            for field in fields:
                del self[field]
        return deserialized_successfully

    def serialize(self):
        if not ytcfg.get("yt", "serialize"):
            return
        self._mrep.store(self.ds.parameter_filename + ".yt")

    def _get_tree(self, nvals):
        xax = self.ds.coordinates.x_axis[self.axis]
        yax = self.ds.coordinates.y_axis[self.axis]
        xd = self.ds.domain_dimensions[xax]
        yd = self.ds.domain_dimensions[yax]
        bounds = (
            self.ds.domain_left_edge[xax],
            self.ds.domain_right_edge[xax],
            self.ds.domain_left_edge[yax],
            self.ds.domain_right_edge[yax],
        )
        return QuadTree(
            np.array([xd, yd], dtype="int64"), nvals, bounds, method=self.method
        )

    def _initialize_chunk(self, chunk, tree):
        icoords = chunk.icoords
        xax = self.ds.coordinates.x_axis[self.axis]
        yax = self.ds.coordinates.y_axis[self.axis]
        i1 = icoords[:, xax]
        i2 = icoords[:, yax]
        ilevel = chunk.ires * self.ds.ires_factor
        tree.initialize_chunk(i1, i2, ilevel)

    def _handle_chunk(self, chunk, fields, tree):
        mylog.debug(
            "Adding chunk (%s) to tree (%0.3e GB RAM)",
            chunk.ires.size,
            get_memory_usage() / 1024.0,
        )
        if self.method in ("min", "max") or self._sum_only:
            dl = self.ds.quan(1.0, "")
        else:
            ax_name = self.ds.coordinates.axis_name[self.axis]
            dl = chunk["index", f"path_element_{ax_name}"]
            # This is done for cases where our path element does not have a CGS
            # equivalent.  Once "preferred units" have been implemented, this
            # will not be necessary at all, as the final conversion will occur
            # at the display layer.
            if not dl.units.is_dimensionless:
                dl.convert_to_units(self.ds.unit_system["length"])
        v = np.empty((chunk.ires.size, len(fields)), dtype="float64")
        for i, field in enumerate(fields):
            d = chunk[field] * dl
            v[:, i] = d
        if self.weight_field is not None:
            w = chunk[self.weight_field]
            np.multiply(v, w[:, None], v)
            np.multiply(w, dl, w)
        else:
            w = np.ones(chunk.ires.size, dtype="float64")
        icoords = chunk.icoords
        xax = self.ds.coordinates.x_axis[self.axis]
        yax = self.ds.coordinates.y_axis[self.axis]
        i1 = icoords[:, xax]
        i2 = icoords[:, yax]
        ilevel = chunk.ires * self.ds.ires_factor
        tree.add_chunk_to_tree(i1, i2, ilevel, v, w)


class YTCoveringGrid(YTSelectionContainer3D):
    """A 3D region with all data extracted to a single, specified
    resolution.  Left edge should align with a cell boundary, but
    defaults to the closest cell boundary.

    Parameters
    ----------
    level : int
        The resolution level data to which data will be gridded. Level
        0 is the root grid dx for that dataset.
    left_edge : array_like
        The left edge of the region to be extracted.  Specify units by supplying
        a YTArray, otherwise code length units are assumed.
    dims : array_like
        Number of cells along each axis of resulting covering_grid
    fields : array_like, optional
        A list of fields that you'd like pre-generated for your object
    num_ghost_zones : integer, optional
        The number of padding ghost zones used when accessing fields.
    data_source :
        An existing data object to intersect with the covering grid. Grid points
        outside the data_source will exist as empty values.

    Examples
    --------
    >>> cube = ds.covering_grid(2, left_edge=[0.0, 0.0, 0.0], dims=[128, 128, 128])
    """

    _spatial = True
    _type_name = "covering_grid"
    _con_args = ("level", "left_edge", "ActiveDimensions")
    _container_fields = (
        ("index", "dx"),
        ("index", "dy"),
        ("index", "dz"),
        ("index", "x"),
        ("index", "y"),
        ("index", "z"),
    )
    _base_grid = None

    def __init__(
        self,
        level,
        left_edge,
        dims,
        fields=None,
        ds=None,
        num_ghost_zones=0,
        use_pbar=True,
        field_parameters=None,
        *,
        data_source=None,
    ):
        if field_parameters is None:
            center = None
        else:
            center = field_parameters.get("center", None)
        super().__init__(center, ds, field_parameters, data_source=data_source)

        self.level = level
        self.left_edge = self._sanitize_edge(left_edge)
        self.ActiveDimensions = self._sanitize_dims(dims)

        rdx = self.ds.domain_dimensions * self.ds.relative_refinement(0, level)

        # normalize dims as a non-zero dim array
        dims = np.array(list(always_iterable(dims)))
        rdx[np.where(dims - 2 * num_ghost_zones <= 1)] = 1  # issue 602
        self.base_dds = self.ds.domain_width / self.ds.domain_dimensions
        self.dds = self.ds.domain_width / rdx.astype("float64")
        self.right_edge = self.left_edge + self.ActiveDimensions * self.dds
        self._num_ghost_zones = num_ghost_zones
        self._use_pbar = use_pbar
        self.global_startindex = (
            np.rint((self.left_edge - self.ds.domain_left_edge) / self.dds).astype(
                "int64"
            )
            + self.ds.domain_offset
        )
        self._setup_data_source()
        self.get_data(fields)

    def get_global_startindex(self):
        r"""Get the global start index of the covering grid."""
        return self.global_startindex

    def to_xarray(self, fields=None):
        r"""Export this fixed-resolution object to an xarray Dataset

        This function will take a regularized grid and optionally a list of
        fields and return an xarray Dataset object.  If xarray is not
        importable, this will raise ImportError.

        Parameters
        ----------
        fields : list of strings or tuple field names, default None
            If this is supplied, it is the list of fields to be exported into
            the data frame.  If not supplied, whatever fields presently exist
            will be used.

        Returns
        -------
        arr : Dataset
            The data contained in the object.

        Examples
        --------

        >>> dd = ds.r[::256j, ::256j, ::256j]
        >>> xf1 = dd.to_xarray([("gas", "density"), ("gas", "temperature")])
        >>> dd[("gas", "velocity_magnitude")]
        >>> xf2 = dd.to_xarray()
        """
        import xarray as xr

        data = {}
        coords = {}
        for f in fields or self.field_data.keys():
            data[f] = {
                "dims": (
                    "x",
                    "y",
                    "z",
                ),
                "data": self[f],
                "attrs": {"units": str(self[f].uq)},
            }
        # We have our data, so now we generate both our coordinates and our metadata.
        LE = self.LeftEdge + self.dds / 2.0
        RE = self.RightEdge - self.dds / 2.0
        N = self.ActiveDimensions
        u = str(LE.uq)
        for i, ax in enumerate("xyz"):
            coords[ax] = {
                "dims": (ax,),
                "data": np.mgrid[LE[i] : RE[i] : N[i] * 1j],
                "attrs": {"units": u},
            }
        return xr.Dataset.from_dict({"data_vars": data, "coords": coords})

    @property
    def icoords(self):
        ic = np.indices(self.ActiveDimensions).astype("int64")
        return np.column_stack(
            [i.ravel() + gi for i, gi in zip(ic, self.get_global_startindex())]
        )

    @property
    def fwidth(self):
        fw = np.ones((self.ActiveDimensions.prod(), 3), dtype="float64")
        fw *= self.dds
        return fw

    @property
    def fcoords(self):
        LE = self.LeftEdge + self.dds / 2.0
        RE = self.RightEdge - self.dds / 2.0
        N = self.ActiveDimensions
        fc = np.mgrid[
            LE[0] : RE[0] : N[0] * 1j,
            LE[1] : RE[1] : N[1] * 1j,
            LE[2] : RE[2] : N[2] * 1j,
        ]
        return np.column_stack([f.ravel() for f in fc])

    @property
    def ires(self):
        tr = np.ones(self.ActiveDimensions.prod(), dtype="int64")
        tr *= self.level
        return tr

    def set_field_parameter(self, name, val):
        super().set_field_parameter(name, val)
        if self._data_source is not None:
            self._data_source.set_field_parameter(name, val)

    def _sanitize_dims(self, dims):
        if not is_sequence(dims):
            dims = [dims] * len(self.ds.domain_left_edge)
        if len(dims) != len(self.ds.domain_left_edge):
            raise RuntimeError(
                "Length of dims must match the dimensionality of the dataset"
            )
        return np.array(dims, dtype="int32")

    def _sanitize_edge(self, edge):
        if not is_sequence(edge):
            edge = [edge] * len(self.ds.domain_left_edge)
        if len(edge) != len(self.ds.domain_left_edge):
            raise RuntimeError(
                "Length of edges must match the dimensionality of the dataset"
            )
        if hasattr(edge, "units"):
            if edge.units.registry is self.ds.unit_registry:
                return edge
            edge_units = edge.units.copy()
            edge_units.registry = self.ds.unit_registry
        else:
            edge_units = "code_length"
        return self.ds.arr(edge, edge_units, dtype="float64")

    def _reshape_vals(self, arr):
        if len(arr.shape) == 3:
            return arr
        return arr.reshape(self.ActiveDimensions, order="C")

    @property
    def shape(self):
        return tuple(self.ActiveDimensions.tolist())

    def _setup_data_source(self):

        reg = self.ds.region(self.center, self.left_edge, self.right_edge)
        if self._data_source is None:
            # note: https://github.com/yt-project/yt/pull/4063 implemented
            # a data_source kwarg for YTCoveringGrid, but not YTArbitraryGrid
            # so as of 4063, this will always be True for YTArbitraryGrid
            # instances.
            self._data_source = reg
        else:
            self._data_source = self.ds.intersection([self._data_source, reg])

        self._data_source.min_level = 0
        self._data_source.max_level = self.level
        # This triggers "special" behavior in the RegionSelector to ensure we
        # select *cells* whose bounding boxes overlap with our region, not just
        # their cell centers.
        self._data_source.loose_selection = True

    def get_data(self, fields=None):
        if fields is None:
            return
        fields = self._determine_fields(fields)
        fields_to_get = [f for f in fields if f not in self.field_data]
        fields_to_get = self._identify_dependencies(fields_to_get)
        if len(fields_to_get) == 0:
            return
        try:
            fill, gen, part, alias = self._split_fields(fields_to_get)
        except NeedsGridType as e:
            if self._num_ghost_zones == 0:
                raise RuntimeError(
                    "Attempting to access a field that needs ghost zones, but "
                    "num_ghost_zones = %s. You should create the covering grid "
                    "with nonzero num_ghost_zones." % self._num_ghost_zones
                ) from e
            else:
                raise

        # checking if we have a sph particles
        if len(part) == 0:
            is_sph_field = False
        else:
            is_sph_field = self.ds.field_info[part[0]].is_sph_field

        if len(part) > 0 and len(alias) == 0:
            if is_sph_field:
                self._fill_sph_particles(fields)
                for field in fields:
                    if field in gen:
                        gen.remove(field)
            else:
                self._fill_particles(part)

        if len(fill) > 0:
            self._fill_fields(fill)
        for a, f in sorted(alias.items()):
            if f.sampling_type == "particle" and not is_sph_field:
                self[a] = self._data_source[f]
            else:
                self[a] = f(self)
            self.field_data[a].convert_to_units(f.output_units)

        if len(gen) > 0:
            part_gen = []
            cell_gen = []
            for field in gen:
                finfo = self.ds.field_info[field]
                if finfo.sampling_type == "particle":
                    part_gen.append(field)
                else:
                    cell_gen.append(field)
            self._generate_fields(cell_gen)
            for p in part_gen:
                self[p] = self._data_source[p]

    def _split_fields(self, fields_to_get):
        fill, gen = self.index._split_fields(fields_to_get)
        particles = []
        alias = {}
        for field in gen:
            finfo = self.ds._get_field_info(*field)
            if finfo.is_alias:
                alias[field] = finfo
                continue
            try:
                finfo.check_available(self)
            except NeedsOriginalGrid:
                fill.append(field)
        for field in fill:
            finfo = self.ds._get_field_info(*field)
            if finfo.sampling_type == "particle":
                particles.append(field)
        gen = [f for f in gen if f not in fill and f not in alias]
        fill = [f for f in fill if f not in particles]
        return fill, gen, particles, alias

    def _fill_particles(self, part):
        for p in part:
            self[p] = self._data_source[p]

    def _fill_sph_particles(self, fields):
        # checks that we have the field and gets information
        fields = [f for f in fields if f not in self.field_data]
        if len(fields) == 0:
            return

        smoothing_style = getattr(self.ds, "sph_smoothing_style", "scatter")
        normalize = getattr(self.ds, "use_sph_normalization", True)

        bounds, size = self._get_grid_bounds_size()

        period = self.ds.coordinates.period.copy()
        if hasattr(period, "in_units"):
            period = period.in_units("code_length").d
        # TODO maybe there is a better way of handling this
        is_periodic = int(any(self.ds.periodicity))

        if smoothing_style == "scatter":
            for field in fields:
                fi = self.ds._get_field_info(field)
                ptype = fi.name[0]
                if ptype not in self.ds._sph_ptypes:
                    raise KeyError(f"{ptype} is not a SPH particle type!")
                buff = np.zeros(size, dtype="float64")
                if normalize:
                    buff_den = np.zeros(size, dtype="float64")

                pbar = tqdm(desc=f"Interpolating SPH field {field}")
                for chunk in self._data_source.chunks([field], "io"):
                    px = chunk[(ptype, "particle_position_x")].in_base("code").d
                    py = chunk[(ptype, "particle_position_y")].in_base("code").d
                    pz = chunk[(ptype, "particle_position_z")].in_base("code").d
                    hsml = chunk[(ptype, "smoothing_length")].in_base("code").d
                    mass = chunk[(ptype, "particle_mass")].in_base("code").d
                    dens = chunk[(ptype, "density")].in_base("code").d
                    field_quantity = chunk[field].d

                    pixelize_sph_kernel_arbitrary_grid(
                        buff,
                        px,
                        py,
                        pz,
                        hsml,
                        mass,
                        dens,
                        field_quantity,
                        bounds,
                        pbar=pbar,
                        check_period=is_periodic,
                        period=period,
                    )
                    if normalize:
                        pixelize_sph_kernel_arbitrary_grid(
                            buff_den,
                            px,
                            py,
                            pz,
                            hsml,
                            mass,
                            dens,
                            np.ones(dens.shape[0]),
                            bounds,
                            pbar=pbar,
                            check_period=is_periodic,
                            period=period,
                        )

                if normalize:
                    normalization_3d_utility(buff, buff_den)

                self[field] = self.ds.arr(buff, fi.units)
                pbar.close()

        if smoothing_style == "gather":
            num_neighbors = getattr(self.ds, "num_neighbors", 32)
            for field in fields:
                buff = np.zeros(size, dtype="float64")

                fields_to_get = [
                    "particle_position",
                    "density",
                    "particle_mass",
                    "smoothing_length",
                    field[1],
                ]
                all_fields = all_data(self.ds, field[0], fields_to_get, kdtree=True)

                fi = self.ds._get_field_info(field)
                interpolate_sph_grid_gather(
                    buff,
                    all_fields["particle_position"],
                    bounds,
                    all_fields["smoothing_length"],
                    all_fields["particle_mass"],
                    all_fields["density"],
                    all_fields[field[1]].in_units(fi.units),
                    self.ds.index.kdtree,
                    use_normalization=normalize,
                    num_neigh=num_neighbors,
                )

                self[field] = self.ds.arr(buff, fi.units)

    def _fill_fields(self, fields):
        fields = [f for f in fields if f not in self.field_data]
        if len(fields) == 0:
            return
        output_fields = [
            np.zeros(self.ActiveDimensions, dtype="float64") for field in fields
        ]
        domain_dims = self.ds.domain_dimensions.astype(
            "int64"
        ) * self.ds.relative_refinement(0, self.level)
        refine_by = self.ds.refine_by
        if not is_sequence(self.ds.refine_by):
            refine_by = [refine_by, refine_by, refine_by]
        refine_by = np.array(refine_by, dtype="i8")
        for chunk in parallel_objects(self._data_source.chunks(fields, "io")):
            input_fields = [chunk[field] for field in fields]
            # NOTE: This usage of "refine_by" is actually *okay*, because it's
            # being used with respect to iref, which is *already* scaled!
            fill_region(
                input_fields,
                output_fields,
                self.level,
                self.global_startindex,
                chunk.icoords,
                chunk.ires,
                domain_dims,
                refine_by,
            )
        if self.comm.size > 1:
            for i in range(len(fields)):
                output_fields[i] = self.comm.mpi_allreduce(output_fields[i], op="sum")
        for name, v in zip(fields, output_fields):
            fi = self.ds._get_field_info(*name)
            self[name] = self.ds.arr(v, fi.units)

    def _generate_container_field(self, field):
        rv = self.ds.arr(np.ones(self.ActiveDimensions, dtype="float64"), "")
        axis_name = self.ds.coordinates.axis_name
        if field == ("index", f"d{axis_name[0]}"):
            np.multiply(rv, self.dds[0], rv)
        elif field == ("index", f"d{axis_name[1]}"):
            np.multiply(rv, self.dds[1], rv)
        elif field == ("index", f"d{axis_name[2]}"):
            np.multiply(rv, self.dds[2], rv)
        elif field == ("index", axis_name[0]):
            x = np.mgrid[
                self.left_edge[0]
                + 0.5 * self.dds[0] : self.right_edge[0]
                - 0.5 * self.dds[0] : self.ActiveDimensions[0] * 1j
            ]
            np.multiply(rv, x[:, None, None], rv)
        elif field == ("index", axis_name[1]):
            y = np.mgrid[
                self.left_edge[1]
                + 0.5 * self.dds[1] : self.right_edge[1]
                - 0.5 * self.dds[1] : self.ActiveDimensions[1] * 1j
            ]
            np.multiply(rv, y[None, :, None], rv)
        elif field == ("index", axis_name[2]):
            z = np.mgrid[
                self.left_edge[2]
                + 0.5 * self.dds[2] : self.right_edge[2]
                - 0.5 * self.dds[2] : self.ActiveDimensions[2] * 1j
            ]
            np.multiply(rv, z[None, None, :], rv)
        else:
            raise KeyError(field)
        return rv

    @property
    def LeftEdge(self):
        return self.left_edge

    @property
    def RightEdge(self):
        return self.right_edge

    def deposit(self, positions, fields=None, method=None, kernel_name="cubic"):
        cls = getattr(particle_deposit, f"deposit_{method}", None)
        if cls is None:
            raise YTParticleDepositionNotImplemented(method)
        # We allocate number of zones, not number of octs. Everything
        # inside this is Fortran ordered because of the ordering in the
        # octree deposit routines, so we reverse it here to match the
        # convention there
        nvals = tuple(self.ActiveDimensions[::-1])
        # append a dummy dimension because we are only depositing onto
        # one grid
        op = cls(nvals + (1,), kernel_name)
        op.initialize()
        op.process_grid(self, positions, fields)
        # Fortran-ordered, so transpose.
        vals = op.finalize().transpose()
        # squeeze dummy dimension we appended above
        return np.squeeze(vals, axis=0)

    def write_to_gdf(self, gdf_path, fields, nprocs=1, field_units=None, **kwargs):
        r"""
        Write the covering grid data to a GDF file.

        Parameters
        ----------
        gdf_path : string
            Pathname of the GDF file to write.
        fields : list of strings
            Fields to write to the GDF file.
        nprocs : integer, optional
            Split the covering grid into *nprocs* subgrids before
            writing to the GDF file. Default: 1
        field_units : dictionary, optional
            Dictionary of units to convert fields to. If not set, fields are
            in their default units.
        All remaining keyword arguments are passed to
        yt.utilities.grid_data_format.writer.write_to_gdf.

        Examples
        --------
        >>> cube.write_to_gdf(
        ...     "clumps.h5",
        ...     [("gas", "density"), ("gas", "temperature")],
        ...     nprocs=16,
        ...     overwrite=True,
        ... )
        """
        data = {}
        for field in fields:
            if field in field_units:
                units = field_units[field]
            else:
                units = str(self[field].units)
            data[field] = (self[field].in_units(units).v, units)
        le = self.left_edge.v
        re = self.right_edge.v
        bbox = np.array([[l, r] for l, r in zip(le, re)])
        ds = load_uniform_grid(
            data,
            self.ActiveDimensions,
            bbox=bbox,
            length_unit=self.ds.length_unit,
            time_unit=self.ds.time_unit,
            mass_unit=self.ds.mass_unit,
            nprocs=nprocs,
            sim_time=self.ds.current_time.v,
        )
        write_to_gdf(ds, gdf_path, **kwargs)

    def _get_grid_bounds_size(self):
        dd = self.ds.domain_width / 2**self.level
        bounds = np.zeros(6, dtype="float64")

        bounds[0] = self.left_edge[0].in_base("code")
        bounds[1] = bounds[0] + dd[0].d * self.ActiveDimensions[0]
        bounds[2] = self.left_edge[1].in_base("code")
        bounds[3] = bounds[2] + dd[1].d * self.ActiveDimensions[1]
        bounds[4] = self.left_edge[2].in_base("code")
        bounds[5] = bounds[4] + dd[2].d * self.ActiveDimensions[2]
        size = np.ones(3, dtype="int64") * 2**self.level

        return bounds, size

    def to_fits_data(self, fields, length_unit=None):
        r"""Export a set of gridded fields to a FITS file.

        This will export a set of FITS images of either the fields specified
        or all the fields already in the object.

        Parameters
        ----------
        fields : list of strings
            These fields will be pixelized and output. If "None", the keys of the
            FRB will be used.
        length_unit : string, optional
            the length units that the coordinates are written in. The default
            is to use the default length unit of the dataset.
        """
        from yt.visualization.fits_image import FITSImageData

        if length_unit is None:
            length_unit = self.ds.length_unit
        fields = list(iter_fields(fields))
        fid = FITSImageData(self, fields, length_unit=length_unit)
        return fid


class YTArbitraryGrid(YTCoveringGrid):
    """A 3D region with arbitrary bounds and dimensions.

    In contrast to the Covering Grid, this object accepts a left edge, a right
    edge, and dimensions.  This allows it to be used for creating 3D particle
    deposition fields that are independent of the underlying mesh, whether that
    is yt-generated or from the simulation data.  For example, arbitrary boxes
    around particles can be drawn and particle deposition fields can be
    created.  This object will refuse to generate any fluid fields.

    Parameters
    ----------
    left_edge : array_like
        The left edge of the region to be extracted
    right_edge : array_like
        The left edge of the region to be extracted
    dims : array_like
        Number of cells along each axis of resulting grid.

    Examples
    --------
    >>> obj = ds.arbitrary_grid(
    ...     [0.0, 0.0, 0.0], [0.99, 0.99, 0.99], dims=[128, 128, 128]
    ... )
    """

    _spatial = True
    _type_name = "arbitrary_grid"
    _con_args = ("left_edge", "right_edge", "ActiveDimensions")
    _container_fields = (
        ("index", "dx"),
        ("index", "dy"),
        ("index", "dz"),
        ("index", "x"),
        ("index", "y"),
        ("index", "z"),
    )

    def __init__(self, left_edge, right_edge, dims, ds=None, field_parameters=None):
        if field_parameters is None:
            center = None
        else:
            center = field_parameters.get("center", None)
        YTSelectionContainer3D.__init__(self, center, ds, field_parameters)
        self.left_edge = self._sanitize_edge(left_edge)
        self.right_edge = self._sanitize_edge(right_edge)
        self.ActiveDimensions = self._sanitize_dims(dims)
        self.dds = self.base_dds = (
            self.right_edge - self.left_edge
        ) / self.ActiveDimensions
        self.level = 99
        self._setup_data_source()

    def _fill_fields(self, fields):
        fields = [f for f in fields if f not in self.field_data]
        if len(fields) == 0:
            return
        # It may be faster to adapt fill_region_float to fill multiple fields
        # instead of looping here
        for field in fields:
            dest = np.zeros(self.ActiveDimensions, dtype="float64")
            for chunk in self._data_source.chunks(fields, "io"):
                fill_region_float(
                    chunk.fcoords,
                    chunk.fwidth,
                    chunk[field],
                    self.left_edge,
                    self.right_edge,
                    dest,
                    1,
                    self.ds.domain_width,
                    int(any(self.ds.periodicity)),
                )
            fi = self.ds._get_field_info(field)
            self[field] = self.ds.arr(dest, fi.units)

    def _get_grid_bounds_size(self):
        bounds = np.empty(6, dtype="float64")
        bounds[0] = self.left_edge[0].in_base("code")
        bounds[2] = self.left_edge[1].in_base("code")
        bounds[4] = self.left_edge[2].in_base("code")
        bounds[1] = self.right_edge[0].in_base("code")
        bounds[3] = self.right_edge[1].in_base("code")
        bounds[5] = self.right_edge[2].in_base("code")
        size = self.ActiveDimensions

        return bounds, size


class LevelState:
    current_dx = None
    current_dims = None
    current_level = None
    global_startindex = None
    old_global_startindex = None
    fields = None
    data_source = None

    # These are all cached here as numpy arrays, without units, in
    # code_lengths.
    domain_width = None
    domain_left_edge = None
    domain_right_edge = None
    left_edge = None
    right_edge = None
    base_dx = None
    dds = None


class YTSmoothedCoveringGrid(YTCoveringGrid):
    """A 3D region with all data extracted and interpolated to a
    single, specified resolution. (Identical to covering_grid,
    except that it interpolates.)

    Smoothed covering grids start at level 0, interpolating to
    fill the region to level 1, replacing any cells actually
    covered by level 1 data, and then recursively repeating this
    process until it reaches the specified `level`.

    Parameters
    ----------
    level : int
        The resolution level data is uniformly gridded at
    left_edge : array_like
        The left edge of the region to be extracted
    dims : array_like
        Number of cells along each axis of resulting covering_grid.
    fields : array_like, optional
        A list of fields that you'd like pre-generated for your object

    Example
    -------
    cube = ds.smoothed_covering_grid(2, left_edge=[0.0, 0.0, 0.0], \
                              dims=[128, 128, 128])
    """

    _type_name = "smoothed_covering_grid"
    filename = None
    _min_level = None

    @wraps(YTCoveringGrid.__init__)
    def __init__(self, *args, **kwargs):
        ds = kwargs["ds"]
        self._base_dx = (
            ds.domain_right_edge - ds.domain_left_edge
        ) / ds.domain_dimensions.astype("float64")
        self.global_endindex = None
        YTCoveringGrid.__init__(self, *args, **kwargs)
        self._final_start_index = self.global_startindex

    def _setup_data_source(self, level_state=None):
        if level_state is None:
            super()._setup_data_source()
            return
        # We need a buffer region to allow for zones that contribute to the
        # interpolation but are not directly inside our bounds
        level_state.data_source = self.ds.region(
            self.center,
            level_state.left_edge - level_state.current_dx,
            level_state.right_edge + level_state.current_dx,
        )
        level_state.data_source.min_level = level_state.current_level
        level_state.data_source.max_level = level_state.current_level
        self._pdata_source = self.ds.region(
            self.center,
            level_state.left_edge - level_state.current_dx,
            level_state.right_edge + level_state.current_dx,
        )
        self._pdata_source.min_level = level_state.current_level
        self._pdata_source.max_level = level_state.current_level

    def _compute_minimum_level(self):
        # This attempts to determine the minimum level that we should be
        # starting on for this box.  It does this by identifying the minimum
        # level that could contribute to the minimum bounding box at that
        # level; that means that all cells from coarser levels will be replaced.
        if self._min_level is not None:
            return self._min_level
        ils = LevelState()
        min_level = 0
        for l in range(self.level, 0, -1):
            dx = self._base_dx / self.ds.relative_refinement(0, l)
            start_index, end_index, dims = self._minimal_box(dx)
            ils.left_edge = start_index * dx + self.ds.domain_left_edge
            ils.right_edge = ils.left_edge + dx * dims
            ils.current_dx = dx
            ils.current_level = l
            self._setup_data_source(ils)
            # Reset the max_level
            ils.data_source.min_level = 0
            ils.data_source.max_level = l
            ils.data_source.loose_selection = False
            min_level = self.level
            for chunk in ils.data_source.chunks([], "io"):
                # With our odd selection methods, we can sometimes get no-sized ires.
                ir = chunk.ires
                if ir.size == 0:
                    continue
                min_level = min(ir.min(), min_level)
            if min_level >= l:
                break
        self._min_level = min_level
        return min_level

    def _fill_fields(self, fields):
        fields = [f for f in fields if f not in self.field_data]
        if len(fields) == 0:
            return
        ls = self._initialize_level_state(fields)
        min_level = self._compute_minimum_level()
        # NOTE: This usage of "refine_by" is actually *okay*, because it's
        # being used with respect to iref, which is *already* scaled!
        refine_by = self.ds.refine_by
        if not is_sequence(self.ds.refine_by):
            refine_by = [refine_by, refine_by, refine_by]
        refine_by = np.array(refine_by, dtype="i8")

        runtime_errors_count = 0
        for level in range(self.level + 1):
            if level < min_level:
                self._update_level_state(ls)
                continue
            nd = self.ds.dimensionality
            refinement = np.zeros_like(ls.base_dx)
            refinement += self.ds.relative_refinement(0, ls.current_level)
            refinement[nd:] = 1
            domain_dims = self.ds.domain_dimensions * refinement
            domain_dims = domain_dims.astype("int64")
            tot = ls.current_dims.prod()
            for chunk in ls.data_source.chunks(fields, "io"):
                chunk[fields[0]]
                input_fields = [chunk[field] for field in fields]
                tot -= fill_region(
                    input_fields,
                    ls.fields,
                    ls.current_level,
                    ls.global_startindex,
                    chunk.icoords,
                    chunk.ires,
                    domain_dims,
                    refine_by,
                )
            if level == 0 and tot != 0:
                runtime_errors_count += 1
            self._update_level_state(ls)
        if runtime_errors_count:
            warnings.warn(
                "Something went wrong during field computation. "
                "This is likely due to missing ghost-zones support "
                f"in class {type(self.ds)}",
                category=RuntimeWarning,
            )
            mylog.debug("Caught %d runtime errors.", runtime_errors_count)
        for name, v in zip(fields, ls.fields):
            if self.level > 0:
                v = v[1:-1, 1:-1, 1:-1]
            fi = self.ds._get_field_info(*name)
            self[name] = self.ds.arr(v, fi.units)

    def _initialize_level_state(self, fields):
        ls = LevelState()
        ls.domain_width = self.ds.domain_width
        ls.domain_left_edge = self.ds.domain_left_edge
        ls.domain_right_edge = self.ds.domain_right_edge
        ls.base_dx = self._base_dx
        ls.dds = self.dds
        ls.left_edge = self.left_edge
        ls.right_edge = self.right_edge
        for att in (
            "domain_width",
            "domain_left_edge",
            "domain_right_edge",
            "left_edge",
            "right_edge",
            "base_dx",
            "dds",
        ):
            setattr(ls, att, getattr(ls, att).in_units("code_length").d)
        ls.current_dx = ls.base_dx
        ls.current_level = 0
        ls.global_startindex, end_index, idims = self._minimal_box(ls.current_dx)
        ls.current_dims = idims.astype("int32")
        ls.left_edge = ls.global_startindex * ls.current_dx + self.ds.domain_left_edge.d
        ls.right_edge = ls.left_edge + ls.current_dims * ls.current_dx
        ls.fields = [np.zeros(idims, dtype="float64") - 999 for field in fields]
        self._setup_data_source(ls)
        return ls

    def _minimal_box(self, dds):
        LL = self.left_edge.d - self.ds.domain_left_edge.d
        # Nudge in case we're on the edge
        LL += np.finfo(np.float64).eps
        LS = self.right_edge.d - self.ds.domain_left_edge.d
        LS += np.finfo(np.float64).eps
        cell_start = LL / dds  # This is the cell we're inside
        cell_end = LS / dds
        if self.level == 0:
            start_index = np.array(np.floor(cell_start), dtype="int64")
            end_index = np.array(np.ceil(cell_end), dtype="int64")
            dims = np.rint((self.ActiveDimensions * self.dds.d) / dds).astype("int64")
        else:
            # Give us one buffer
            start_index = np.rint(cell_start).astype("int64") - 1
            # How many root cells do we occupy?
            end_index = np.rint(cell_end).astype("int64")
            dims = end_index - start_index + 1
        return start_index, end_index.astype("int64"), dims.astype("int32")

    def _update_level_state(self, level_state):
        ls = level_state
        if ls.current_level >= self.level:
            return
        rf = float(self.ds.relative_refinement(ls.current_level, ls.current_level + 1))
        ls.current_level += 1
        nd = self.ds.dimensionality
        refinement = np.zeros_like(ls.base_dx)
        refinement += self.ds.relative_refinement(0, ls.current_level)
        refinement[nd:] = 1
        ls.current_dx = ls.base_dx / refinement
        ls.old_global_startindex = ls.global_startindex
        ls.global_startindex, end_index, ls.current_dims = self._minimal_box(
            ls.current_dx
        )
        ls.left_edge = ls.global_startindex * ls.current_dx + self.ds.domain_left_edge.d
        ls.right_edge = ls.left_edge + ls.current_dims * ls.current_dx
        input_left = (level_state.old_global_startindex) * rf + 1
        new_fields = []
        for input_field in level_state.fields:
            output_field = np.zeros(ls.current_dims, dtype="float64")
            output_left = level_state.global_startindex + 0.5
            ghost_zone_interpolate(
                rf, input_field, input_left, output_field, output_left
            )
            new_fields.append(output_field)
        level_state.fields = new_fields
        self._setup_data_source(ls)


class YTSurface(YTSelectionContainer3D):
    r"""This surface object identifies isocontours on a cell-by-cell basis,
    with no consideration of global connectedness, and returns the vertices
    of the Triangles in that isocontour.

    This object simply returns the vertices of all the triangles calculated by
    the `marching cubes <https://en.wikipedia.org/wiki/Marching_cubes>`_
    algorithm; for more complex operations, such as identifying connected sets
    of cells above a given threshold, see the extract_connected_sets function.
    This is more useful for calculating, for instance, total isocontour area, or
    visualizing in an external program (such as `MeshLab
    <http://www.meshlab.net>`_.)  The object has the properties .vertices and
    will sample values if a field is requested.  The values are interpolated to
    the center of a given face.

    Parameters
    ----------
    data_source : YTSelectionContainer
        This is the object which will used as a source
    surface_field : string
        Any field that can be obtained in a data object.  This is the field
        which will be isocontoured.
    field_value : float, YTQuantity, or unit tuple
        The value at which the isocontour should be calculated.

    Examples
    --------
    This will create a data object, find a nice value in the center, and
    output the vertices to "triangles.obj" after rescaling them.

    >>> from yt.units import kpc
    >>> sp = ds.sphere("max", (10, "kpc"))
    >>> surf = ds.surface(sp, ("gas", "density"), 5e-27)
    >>> print(surf[("gas", "temperature")])
    >>> print(surf.vertices)
    >>> bounds = [
    ...     (sp.center[i] - 5.0 * kpc, sp.center[i] + 5.0 * kpc) for i in range(3)
    ... ]
    >>> surf.export_ply("my_galaxy.ply", bounds=bounds)
    """
    _type_name = "surface"
    _con_args = ("data_source", "surface_field", "field_value")
    _container_fields = (
        ("index", "dx"),
        ("index", "dy"),
        ("index", "dz"),
        ("index", "x"),
        ("index", "y"),
        ("index", "z"),
    )

    def __init__(self, data_source, surface_field, field_value, ds=None):
        self.data_source = data_source
        self.surface_field = data_source._determine_fields(surface_field)[0]
        finfo = data_source.ds.field_info[self.surface_field]
        try:
            self.field_value = field_value.to(finfo.units)
        except AttributeError:
            if isinstance(field_value, tuple):
                self.field_value = data_source.ds.quan(*field_value)
                self.field_value = self.field_value.to(finfo.units)
            else:
                self.field_value = data_source.ds.quan(field_value, finfo.units)
        self.vertex_samples = YTFieldData()
        center = data_source.get_field_parameter("center")
        super().__init__(center=center, ds=ds)

    def _generate_container_field(self, field):
        self.get_data(field)
        return self[field]

    def get_data(self, fields=None, sample_type="face", no_ghost=False):
        if isinstance(fields, list) and len(fields) > 1:
            for field in fields:
                self.get_data(field)
            return
        elif isinstance(fields, list):
            fields = fields[0]
        # Now we have a "fields" value that is either a string or None
        if fields is not None:
            mylog.info("Extracting (sampling: %s)", fields)
        verts = []
        samples = []
        for _io_chunk in parallel_objects(self.data_source.chunks([], "io")):
            for block, mask in self.data_source.blocks:
                my_verts = self._extract_isocontours_from_grid(
                    block,
                    self.surface_field,
                    self.field_value,
                    mask,
                    fields,
                    sample_type,
                    no_ghost=no_ghost,
                )
                if fields is not None:
                    my_verts, svals = my_verts
                    samples.append(svals)
                verts.append(my_verts)
        verts = np.concatenate(verts).transpose()
        verts = self.comm.par_combine_object(verts, op="cat", datatype="array")
        # verts is an ndarray here and will always be in code units, so we
        # expose it in the public API as a YTArray
        self._vertices = self.ds.arr(verts, "code_length")
        if fields is not None:
            samples = uconcatenate(samples)
            samples = self.comm.par_combine_object(samples, op="cat", datatype="array")
            if sample_type == "face":
                self[fields] = samples
            elif sample_type == "vertex":
                self.vertex_samples[fields] = samples

    def _extract_isocontours_from_grid(
        self,
        grid,
        field,
        value,
        mask,
        sample_values=None,
        sample_type="face",
        no_ghost=False,
    ):
        # TODO: check if multiple fields can be passed here
        vals = grid.get_vertex_centered_data([field], no_ghost=no_ghost)[field]
        if sample_values is not None:
            # TODO: is no_ghost=False correct here?
            svals = grid.get_vertex_centered_data([sample_values])[sample_values]
        else:
            svals = None

        sample_type = {"face": 1, "vertex": 2}[sample_type]
        my_verts = march_cubes_grid(
            value, vals, mask, grid.LeftEdge, grid.dds, svals, sample_type
        )
        return my_verts

    def calculate_flux(self, field_x, field_y, field_z, fluxing_field=None):
        r"""This calculates the flux over the surface.

        This function will conduct `Marching Cubes`_ on all the cells in a
        given data container (grid-by-grid), and then for each identified
        triangular segment of an isocontour in a given cell, calculate the
        gradient (i.e., normal) in the isocontoured field, interpolate the local
        value of the "fluxing" field, the area of the triangle, and then return:

        area * local_flux_value * (n dot v)

        Where area, local_value, and the vector v are interpolated at the barycenter
        (weighted by the vertex values) of the triangle.  Note that this
        specifically allows for the field fluxing across the surface to be
        *different* from the field being contoured.  If the fluxing_field is
        not specified, it is assumed to be 1.0 everywhere, and the raw flux
        with no local-weighting is returned.

        Additionally, the returned flux is defined as flux *into* the surface,
        not flux *out of* the surface.

        Parameters
        ----------
        field_x : string
            The x-component field
        field_y : string
            The y-component field
        field_z : string
            The z-component field
        fluxing_field : string, optional
            The field whose passage over the surface is of interest.  If not
            specified, assumed to be 1.0 everywhere.

        Returns
        -------
        flux : YTQuantity
            The summed flux.

        References
        ----------

        .. _Marching Cubes:
            https://en.wikipedia.org/wiki/Marching_cubes

        Examples
        --------

        This will create a data object, find a nice value in the center, and
        calculate the metal flux over it.

        >>> sp = ds.sphere("max", (10, "kpc"))
        >>> surf = ds.surface(sp, ("gas", "density"), 5e-27)
        >>> flux = surf.calculate_flux(
        ...     ("gas", "velocity_x"),
        ...     ("gas", "velocity_y"),
        ...     ("gas", "velocity_z"),
        ...     ("gas", "metal_density"),
        ... )
        """
        flux = 0.0
        mylog.info("Fluxing %s", fluxing_field)
        for _io_chunk in parallel_objects(self.data_source.chunks([], "io")):
            for block, mask in self.data_source.blocks:
                flux += self._calculate_flux_in_grid(
                    block, mask, field_x, field_y, field_z, fluxing_field
                )
        flux = self.comm.mpi_allreduce(flux, op="sum")
        return flux

    def _calculate_flux_in_grid(
        self, grid, mask, field_x, field_y, field_z, fluxing_field=None
    ):

        vc_fields = [self.surface_field, field_x, field_y, field_z]
        if fluxing_field is not None:
            vc_fields.append(fluxing_field)

        vc_data = grid.get_vertex_centered_data(vc_fields)
        if fluxing_field is None:
            ff = self.ds.arr(
                np.ones_like(vc_data[self.surface_field].d, dtype="float64"),
                "dimensionless",
            )
        else:
            ff = vc_data[fluxing_field]
        surf_vals = vc_data[self.surface_field]
        field_x_vals = vc_data[field_x]
        field_y_vals = vc_data[field_y]
        field_z_vals = vc_data[field_z]
        ret = march_cubes_grid_flux(
            self.field_value,
            surf_vals,
            field_x_vals,
            field_y_vals,
            field_z_vals,
            ff,
            mask,
            grid.LeftEdge,
            grid.dds,
        )
        # assumes all the fluxing fields have the same units
        ret_units = field_x_vals.units * ff.units * grid.dds.units**2
        ret = self.ds.arr(ret, ret_units)
        ret.convert_to_units(self.ds.unit_system[ret_units.dimensions])
        return ret

    _vertices = None

    @property
    def vertices(self):
        if self._vertices is None:
            self.get_data()
        return self._vertices

    @property
    def triangles(self):
        vv = np.empty((self.vertices.shape[1] // 3, 3, 3), dtype="float64")
        vv = self.ds.arr(vv, self.vertices.units)
        for i in range(3):
            for j in range(3):
                vv[:, i, j] = self.vertices[j, i::3]
        return vv

    _surface_area = None

    @property
    def surface_area(self):
        if self._surface_area is not None:
            return self._surface_area
        tris = self.triangles
        x = tris[:, 1, :] - tris[:, 0, :]
        y = tris[:, 2, :] - tris[:, 0, :]
        areas = (x[:, 1] * y[:, 2] - x[:, 2] * y[:, 1]) ** 2
        np.add(areas, (x[:, 2] * y[:, 0] - x[:, 0] * y[:, 2]) ** 2, out=areas)
        np.add(areas, (x[:, 0] * y[:, 1] - x[:, 1] * y[:, 0]) ** 2, out=areas)
        np.sqrt(areas, out=areas)
        self._surface_area = 0.5 * areas.sum()
        return self._surface_area

    def export_obj(
        self,
        filename,
        transparency=1.0,
        dist_fac=None,
        color_field=None,
        emit_field=None,
        color_map=None,
        color_log=True,
        emit_log=True,
        plot_index=None,
        color_field_max=None,
        color_field_min=None,
        emit_field_max=None,
        emit_field_min=None,
    ):
        r"""Export the surface to the OBJ format

        Suitable for visualization in many different programs (e.g., Blender).
        NOTE: this exports an .obj file and an .mtl file, both with the general
        'filename' as a prefix.  The .obj file points to the .mtl file in its
        header, so if you move the 2 files, make sure you change the .obj header
        to account for this. ALSO NOTE: the emit_field needs to be a combination
        of the other 2 fields used to have the emissivity track with the color.

        Parameters
        ----------

        filename : string
            The file this will be exported to.  This cannot be a file-like
            object. If there are no file extensions included - both obj & mtl
            files are created.
        transparency : float
            This gives the transparency of the output surface plot.  Values
            from 0.0 (invisible) to 1.0 (opaque).
        dist_fac : float
            Divide the axes distances by this amount.
        color_field : string
            Should a field be sample and colormapped?
        emit_field : string
            Should we track the emissivity of a field? This should be a
            combination of the other 2 fields being used.
        color_map : string
            Which color map should be applied?
        color_log : bool
            Should the color field be logged before being mapped?
        emit_log : bool
            Should the emitting field be logged before being mapped?
        plot_index : integer
            Index of plot for multiple plots.  If none, then only 1 plot.
        color_field_max : float
            Maximum value of the color field across all surfaces.
        color_field_min : float
            Minimum value of the color field across all surfaces.
        emit_field_max : float
            Maximum value of the emitting field across all surfaces.
        emit_field_min : float
            Minimum value of the emitting field across all surfaces.

        Examples
        --------

        >>> sp = ds.sphere("max", (10, "kpc"))
        >>> trans = 1.0
        >>> surf = ds.surface(sp, ("gas", "density"), 5e-27)
        >>> surf.export_obj("my_galaxy", transparency=trans)

        >>> sp = ds.sphere("max", (10, "kpc"))
        >>> mi, ma = sp.quantities.extrema("temperature")
        >>> rhos = [1e-24, 1e-25]
        >>> trans = [0.5, 1.0]
        >>> for i, r in enumerate(rhos):
        ...     surf = ds.surface(sp, "density", r)
        ...     surf.export_obj(
        ...         "my_galaxy",
        ...         transparency=trans[i],
        ...         color_field="temperature",
        ...         plot_index=i,
        ...         color_field_max=ma,
        ...         color_field_min=mi,
        ...     )

        >>> sp = ds.sphere("max", (10, "kpc"))
        >>> rhos = [1e-24, 1e-25]
        >>> trans = [0.5, 1.0]
        >>> def _Emissivity(field, data):
        ...     return (
        ...         data[("gas", "density")]
        ...         * data[("gas", "density")]
        ...         * np.sqrt(data[("gas", "temperature")])
        ...     )
        >>> ds.add_field(
        ...     ("gas", "emissivity"),
        ...     function=_Emissivity,
        ...     sampling_type="cell",
        ...     units=r"g**2*sqrt(K)/cm**6",
        ... )
        >>> for i, r in enumerate(rhos):
        ...     surf = ds.surface(sp, "density", r)
        ...     surf.export_obj(
        ...         "my_galaxy",
        ...         transparency=trans[i],
        ...         color_field="temperature",
        ...         emit_field="emissivity",
        ...         plot_index=i,
        ...     )

        """
        if color_map is None:
            color_map = ytcfg.get("yt", "default_colormap")
        if self.vertices is None:
            if color_field is not None:
                self.get_data(color_field, "face")
        elif color_field is not None:
            if color_field not in self.field_data:
                self[color_field]
        if color_field is None:
            self.get_data(self.surface_field, "face")
        if emit_field is not None:
            if color_field not in self.field_data:
                self[emit_field]
        only_on_root(
            self._export_obj,
            filename,
            transparency,
            dist_fac,
            color_field,
            emit_field,
            color_map,
            color_log,
            emit_log,
            plot_index,
            color_field_max,
            color_field_min,
            emit_field_max,
            emit_field_min,
        )

    def _color_samples_obj(
        self,
        cs,
        em,
        color_log,
        emit_log,
        color_map,
        arr,
        color_field_max,
        color_field_min,
        color_field,
        emit_field_max,
        emit_field_min,
        emit_field,
    ):  # this now holds for obj files
        if color_field is not None:
            if color_log:
                cs = np.log10(cs)
        if emit_field is not None:
            if emit_log:
                em = np.log10(em)
        if color_field is not None:
            if color_field_min is None:
                cs = [float(field) for field in cs]
                cs = np.array(cs)
                mi = cs.min()
            else:
                mi = color_field_min
                if color_log:
                    mi = np.log10(mi)
            if color_field_max is None:
                cs = [float(field) for field in cs]
                cs = np.array(cs)
                ma = cs.max()
            else:
                ma = color_field_max
                if color_log:
                    ma = np.log10(ma)
            cs = (cs - mi) / (ma - mi)
        else:
            cs[:] = 1.0
        # to get color indices for OBJ formatting
        lut = get_colormap_lut(color_map)

        x = np.mgrid[0.0 : 1.0 : lut[0].shape[0] * 1j]
        arr["cind"][:] = (np.interp(cs, x, x) * (lut[0].shape[0] - 1)).astype("uint8")
        # now, get emission
        if emit_field is not None:
            if emit_field_min is None:
                em = [float(field) for field in em]
                em = np.array(em)
                emi = em.min()
            else:
                emi = emit_field_min
                if emit_log:
                    emi = np.log10(emi)
            if emit_field_max is None:
                em = [float(field) for field in em]
                em = np.array(em)
                ema = em.max()
            else:
                ema = emit_field_max
                if emit_log:
                    ema = np.log10(ema)
            em = (em - emi) / (ema - emi)
            x = np.mgrid[0.0:255.0:2j]  # assume 1 emissivity per color
            arr["emit"][:] = (
                np.interp(em, x, x)
            ) * 2.0  # for some reason, max emiss = 2
        else:
            arr["emit"][:] = 0.0

    @parallel_root_only
    def _export_obj(
        self,
        filename,
        transparency,
        dist_fac=None,
        color_field=None,
        emit_field=None,
        color_map=None,
        color_log=True,
        emit_log=True,
        plot_index=None,
        color_field_max=None,
        color_field_min=None,
        emit_field_max=None,
        emit_field_min=None,
    ):
        if color_map is None:
            color_map = ytcfg.get("yt", "default_colormap")
        if plot_index is None:
            plot_index = 0
        if isinstance(filename, io.IOBase):
            fobj = filename + ".obj"
            fmtl = filename + ".mtl"
        else:
            if plot_index == 0:
                fobj = open(filename + ".obj", "w")
                fmtl = open(filename + ".mtl", "w")
                cc = 1
            else:
                # read in last vertex
                linesave = ""
                for line in fileinput.input(filename + ".obj"):
                    if line[0] == "f":
                        linesave = line
                p = [m.start() for m in finditer(" ", linesave)]
                cc = int(linesave[p[len(p) - 1] :]) + 1
                fobj = open(filename + ".obj", "a")
                fmtl = open(filename + ".mtl", "a")
        ftype = [("cind", "uint8"), ("emit", "float")]
        vtype = [("x", "float"), ("y", "float"), ("z", "float")]
        if plot_index == 0:
            fobj.write("# yt OBJ file\n")
            fobj.write("# www.yt-project.org\n")
            fobj.write(
                f"mtllib {filename}.mtl\n\n"
            )  # use this material file for the faces
            fmtl.write("# yt MLT file\n")
            fmtl.write("# www.yt-project.org\n\n")
        # (0) formulate vertices
        nv = self.vertices.shape[1]  # number of groups of vertices
        f = np.empty(
            nv // self.vertices.shape[0], dtype=ftype
        )  # store sets of face colors
        v = np.empty(nv, dtype=vtype)  # stores vertices
        if color_field is not None:
            cs = self[color_field]
        else:
            cs = np.empty(self.vertices.shape[1] // self.vertices.shape[0])
        if emit_field is not None:
            em = self[emit_field]
        else:
            em = np.empty(self.vertices.shape[1] // self.vertices.shape[0])
        self._color_samples_obj(
            cs,
            em,
            color_log,
            emit_log,
            color_map,
            f,
            color_field_max,
            color_field_min,
            color_field,
            emit_field_max,
            emit_field_min,
            emit_field,
        )  # map color values to color scheme

        lut = get_colormap_lut(color_map)

        # interpolate emissivity to enumerated colors
        emiss = np.interp(
            np.mgrid[0 : lut[0].shape[0]], np.mgrid[0 : len(cs)], f["emit"][:]
        )
        if dist_fac is None:  # then normalize by bounds
            DLE = self.pf.domain_left_edge
            DRE = self.pf.domain_right_edge
            bounds = [(DLE[i], DRE[i]) for i in range(3)]
            for i, ax in enumerate("xyz"):
                # Do the bounds first since we cast to f32
                tmp = self.vertices[i, :]
                np.subtract(tmp, bounds[i][0], tmp)
                w = bounds[i][1] - bounds[i][0]
                np.divide(tmp, w, tmp)
                np.subtract(tmp, 0.5, tmp)  # Center at origin.
                v[ax][:] = tmp
        else:
            for i, ax in enumerate("xyz"):
                tmp = self.vertices[i, :]
                np.divide(tmp, dist_fac, tmp)
                v[ax][:] = tmp
        # (1) write all colors per surface to mtl file
        for i in range(0, lut[0].shape[0]):
            omname = f"material_{i}_{plot_index}"  # name of the material
            fmtl.write(
                f"newmtl {omname}\n"
            )  # the specific material (color) for this face
            fmtl.write(f"Ka {0.0:.6f} {0.0:.6f} {0.0:.6f}\n")  # ambient color, keep off
            fmtl.write(
                f"Kd {lut[0][i]:.6f} {lut[1][i]:.6f} {lut[2][i]:.6f}\n"
            )  # color of face
            fmtl.write(
                f"Ks {0.0:.6f} {0.0:.6f} {0.0:.6f}\n"
            )  # specular color, keep off
            fmtl.write(f"d {transparency:.6f}\n")  # transparency
            fmtl.write(f"em {emiss[i]:.6f}\n")  # emissivity per color
            fmtl.write("illum 2\n")  # not relevant, 2 means highlights on?
            fmtl.write(f"Ns {0:.6f}\n\n")  # keep off, some other specular thing
        # (2) write vertices
        for i in range(0, self.vertices.shape[1]):
            fobj.write(f"v {v['x'][i]:.6f} {v['y'][i]:.6f} {v['z'][i]:.6f}\n")
        fobj.write("#done defining vertices\n\n")
        # (3) define faces and materials for each face
        for i in range(0, self.triangles.shape[0]):
            omname = f"material_{f['cind'][i]}_{plot_index}"  # which color to use
            fobj.write(
                f"usemtl {omname}\n"
            )  # which material to use for this face (color)
            fobj.write(f"f {cc} {cc+1} {cc+2}\n\n")  # vertices to color
            cc = cc + 3
        fmtl.close()
        fobj.close()

    def export_blender(
        self,
        transparency=1.0,
        dist_fac=None,
        color_field=None,
        emit_field=None,
        color_map=None,
        color_log=True,
        emit_log=True,
        plot_index=None,
        color_field_max=None,
        color_field_min=None,
        emit_field_max=None,
        emit_field_min=None,
    ):
        r"""This exports the surface to the OBJ format, suitable for visualization
        in many different programs (e.g., Blender).  NOTE: this exports an .obj file
        and an .mtl file, both with the general 'filename' as a prefix.
        The .obj file points to the .mtl file in its header, so if you move the 2
        files, make sure you change the .obj header to account for this. ALSO NOTE:
        the emit_field needs to be a combination of the other 2 fields used to
        have the emissivity track with the color.

        Parameters
        ----------
        transparency : float
            This gives the transparency of the output surface plot.  Values
            from 0.0 (invisible) to 1.0 (opaque).
        dist_fac : float
            Divide the axes distances by this amount.
        color_field : string
            Should a field be sample and colormapped?
        emit_field : string
            Should we track the emissivity of a field?
              NOTE: this should be a combination of the other 2 fields being used.
        color_map : string
            Which color map should be applied?
        color_log : bool
            Should the color field be logged before being mapped?
        emit_log : bool
            Should the emitting field be logged before being mapped?
        plot_index : integer
            Index of plot for multiple plots.  If none, then only 1 plot.
        color_field_max : float
            Maximum value of the color field across all surfaces.
        color_field_min : float
            Minimum value of the color field across all surfaces.
        emit_field_max : float
            Maximum value of the emitting field across all surfaces.
        emit_field_min : float
            Minimum value of the emitting field across all surfaces.

        Examples
        --------

        >>> sp = ds.sphere("max", (10, "kpc"))
        >>> trans = 1.0
        >>> surf = ds.surface(sp, ("gas", "density"), 5e-27)
        >>> surf.export_obj("my_galaxy", transparency=trans)

        >>> sp = ds.sphere("max", (10, "kpc"))
        >>> mi, ma = sp.quantities.extrema("temperature")[0]
        >>> rhos = [1e-24, 1e-25]
        >>> trans = [0.5, 1.0]
        >>> for i, r in enumerate(rhos):
        ...     surf = ds.surface(sp, "density", r)
        ...     surf.export_obj(
        ...         "my_galaxy",
        ...         transparency=trans[i],
        ...         color_field="temperature",
        ...         plot_index=i,
        ...         color_field_max=ma,
        ...         color_field_min=mi,
        ...     )

        >>> sp = ds.sphere("max", (10, "kpc"))
        >>> rhos = [1e-24, 1e-25]
        >>> trans = [0.5, 1.0]
        >>> def _Emissivity(field, data):
        ...     return (
        ...         data[("gas", "density")]
        ...         * data[("gas", "density")]
        ...         * np.sqrt(data[("gas", "temperature")])
        ...     )
        >>> ds.add_field(("gas", "emissivity"), function=_Emissivity, units="g / cm**6")
        >>> for i, r in enumerate(rhos):
        ...     surf = ds.surface(sp, "density", r)
        ...     surf.export_obj(
        ...         "my_galaxy",
        ...         transparency=trans[i],
        ...         color_field="temperature",
        ...         emit_field="emissivity",
        ...         plot_index=i,
        ...     )

        """
        if color_map is None:
            color_map = ytcfg.get("yt", "default_colormap")
        if self.vertices is None:
            if color_field is not None:
                self.get_data(color_field, "face")
        elif color_field is not None:
            if color_field not in self.field_data:
                self[color_field]
        if color_field is None:
            self.get_data(self.surface_field, "face")
        if emit_field is not None:
            if color_field not in self.field_data:
                self[emit_field]
        fullverts, colors, alpha, emisses, colorindex = only_on_root(
            self._export_blender,
            transparency,
            dist_fac,
            color_field,
            emit_field,
            color_map,
            color_log,
            emit_log,
            plot_index,
            color_field_max,
            color_field_min,
            emit_field_max,
            emit_field_min,
        )
        return fullverts, colors, alpha, emisses, colorindex

    def _export_blender(
        self,
        transparency,
        dist_fac=None,
        color_field=None,
        emit_field=None,
        color_map=None,
        color_log=True,
        emit_log=True,
        plot_index=None,
        color_field_max=None,
        color_field_min=None,
        emit_field_max=None,
        emit_field_min=None,
    ):
        if color_map is None:
            color_map = ytcfg.get("yt", "default_colormap")
        if plot_index is None:
            plot_index = 0
        ftype = [("cind", "uint8"), ("emit", "float")]
        vtype = [("x", "float"), ("y", "float"), ("z", "float")]
        # (0) formulate vertices
        nv = self.vertices.shape[1]  # number of groups of vertices
        f = np.empty(
            nv / self.vertices.shape[0], dtype=ftype
        )  # store sets of face colors
        v = np.empty(nv, dtype=vtype)  # stores vertices
        if color_field is not None:
            cs = self[color_field]
        else:
            cs = np.empty(self.vertices.shape[1] / self.vertices.shape[0])
        if emit_field is not None:
            em = self[emit_field]
        else:
            em = np.empty(self.vertices.shape[1] / self.vertices.shape[0])
        self._color_samples_obj(
            cs,
            em,
            color_log,
            emit_log,
            color_map,
            f,
            color_field_max,
            color_field_min,
            color_field,
            emit_field_max,
            emit_field_min,
            emit_field,
        )  # map color values to color scheme

        lut = get_colormap_lut(color_map)

        # interpolate emissivity to enumerated colors
        emiss = np.interp(
            np.mgrid[0 : lut[0].shape[0]], np.mgrid[0 : len(cs)], f["emit"][:]
        )
        if dist_fac is None:  # then normalize by bounds
            DLE = self.ds.domain_left_edge
            DRE = self.ds.domain_right_edge
            bounds = [(DLE[i], DRE[i]) for i in range(3)]
            for i, ax in enumerate("xyz"):
                # Do the bounds first since we cast to f32
                tmp = self.vertices[i, :]
                np.subtract(tmp, bounds[i][0], tmp)
                w = bounds[i][1] - bounds[i][0]
                np.divide(tmp, w, tmp)
                np.subtract(tmp, 0.5, tmp)  # Center at origin.
                v[ax][:] = tmp
        else:
            for i, ax in enumerate("xyz"):
                tmp = self.vertices[i, :]
                np.divide(tmp, dist_fac, tmp)
                v[ax][:] = tmp
        return v, lut, transparency, emiss, f["cind"]

    def export_ply(
        self,
        filename,
        bounds=None,
        color_field=None,
        color_map=None,
        color_log=True,
        sample_type="face",
        no_ghost=False,
    ):
        r"""This exports the surface to the PLY format, suitable for visualization
        in many different programs (e.g., MeshLab).

        Parameters
        ----------
        filename : string
            The file this will be exported to.  This cannot be a file-like object.
        bounds : list of tuples
            The bounds the vertices will be normalized to.  This is of the format:
            [(xmin, xmax), (ymin, ymax), (zmin, zmax)].  Defaults to the full
            domain.
        color_field : string
            Should a field be sample and colormapped?
        color_map : string
            Which color map should be applied?
        color_log : bool
            Should the color field be logged before being mapped?

        Examples
        --------

        >>> from yt.units import kpc
        >>> sp = ds.sphere("max", (10, "kpc"))
        >>> surf = ds.surface(sp, ("gas", "density"), 5e-27)
        >>> print(surf[("gas", "temperature")])
        >>> print(surf.vertices)
        >>> bounds = [
        ...     (sp.center[i] - 5.0 * kpc, sp.center[i] + 5.0 * kpc) for i in range(3)
        ... ]
        >>> surf.export_ply("my_galaxy.ply", bounds=bounds)
        """
        if color_map is None:
            color_map = ytcfg.get("yt", "default_colormap")
        if self.vertices is None:
            self.get_data(color_field, sample_type, no_ghost=no_ghost)
        elif color_field is not None:
            if sample_type == "face" and color_field not in self.field_data:
                self[color_field]
            elif sample_type == "vertex" and color_field not in self.vertex_samples:
                self.get_data(color_field, sample_type, no_ghost=no_ghost)
        self._export_ply(
            filename, bounds, color_field, color_map, color_log, sample_type
        )

    def _color_samples(self, cs, color_log, color_map, arr):
        if color_log:
            cs = np.log10(cs)
        mi, ma = cs.min(), cs.max()
        cs = (cs - mi) / (ma - mi)
        from yt.visualization.image_writer import map_to_colors

        cs = map_to_colors(cs, color_map)
        arr["red"][:] = cs[0, :, 0]
        arr["green"][:] = cs[0, :, 1]
        arr["blue"][:] = cs[0, :, 2]

    @parallel_root_only
    def _export_ply(
        self,
        filename,
        bounds=None,
        color_field=None,
        color_map=None,
        color_log=True,
        sample_type="face",
    ):
        if color_map is None:
            color_map = ytcfg.get("yt", "default_colormap")
        if hasattr(filename, "read"):
            f = filename
        else:
            f = open(filename, "wb")
        if bounds is None:
            DLE = self.ds.domain_left_edge
            DRE = self.ds.domain_right_edge
            bounds = [(DLE[i], DRE[i]) for i in range(3)]
        elif any([not all([isinstance(be, YTArray) for be in b]) for b in bounds]):
            bounds = [
                tuple(
                    be if isinstance(be, YTArray) else self.ds.quan(be, "code_length")
                    for be in b
                )
                for b in bounds
            ]
        nv = self.vertices.shape[1]
        vs = [
            ("x", "<f"),
            ("y", "<f"),
            ("z", "<f"),
            ("red", "uint8"),
            ("green", "uint8"),
            ("blue", "uint8"),
        ]
        fs = [
            ("ni", "uint8"),
            ("v1", "<i4"),
            ("v2", "<i4"),
            ("v3", "<i4"),
            ("red", "uint8"),
            ("green", "uint8"),
            ("blue", "uint8"),
        ]
        f.write(b"ply\n")
        f.write(b"format binary_little_endian 1.0\n")
        line = "element vertex %i\n" % (nv)
        f.write(line.encode("latin-1"))
        f.write(b"property float x\n")
        f.write(b"property float y\n")
        f.write(b"property float z\n")
        if color_field is not None and sample_type == "vertex":
            f.write(b"property uchar red\n")
            f.write(b"property uchar green\n")
            f.write(b"property uchar blue\n")
            v = np.empty(self.vertices.shape[1], dtype=vs)
            cs = self.vertex_samples[color_field]
            self._color_samples(cs, color_log, color_map, v)
        else:
            v = np.empty(self.vertices.shape[1], dtype=vs[:3])
        line = "element face %i\n" % (nv / 3)
        f.write(line.encode("latin-1"))
        f.write(b"property list uchar int vertex_indices\n")
        if color_field is not None and sample_type == "face":
            f.write(b"property uchar red\n")
            f.write(b"property uchar green\n")
            f.write(b"property uchar blue\n")
            # Now we get our samples
            cs = self[color_field]
            arr = np.empty(cs.shape[0], dtype=np.dtype(fs))
            self._color_samples(cs, color_log, color_map, arr)
        else:
            arr = np.empty(nv // 3, np.dtype(fs[:-3]))
        for i, ax in enumerate("xyz"):
            # Do the bounds first since we cast to f32
            tmp = self.vertices[i, :]
            np.subtract(tmp, bounds[i][0], tmp)
            w = bounds[i][1] - bounds[i][0]
            np.divide(tmp, w, tmp)
            np.subtract(tmp, 0.5, tmp)  # Center at origin.
            v[ax][:] = tmp
        f.write(b"end_header\n")
        v.tofile(f)
        arr["ni"][:] = 3
        vi = np.arange(nv, dtype="<i")
        vi.shape = (nv // 3, 3)
        arr["v1"][:] = vi[:, 0]
        arr["v2"][:] = vi[:, 1]
        arr["v3"][:] = vi[:, 2]
        arr.tofile(f)
        if filename is not f:
            f.close()

    def export_sketchfab(
        self,
        title,
        description,
        api_key=None,
        color_field=None,
        color_map=None,
        color_log=True,
        bounds=None,
        no_ghost=False,
    ):
        r"""This exports Surfaces to SketchFab.com, where they can be viewed
        interactively in a web browser.

        SketchFab.com is a proprietary web service that provides WebGL
        rendering of models.  This routine will use temporary files to
        construct a compressed binary representation (in .PLY format) of the
        Surface and any optional fields you specify and upload it to
        SketchFab.com.  It requires an API key, which can be found on your
        SketchFab.com dashboard.  You can either supply the API key to this
        routine directly or you can place it in the variable
        "sketchfab_api_key" in your ~/.config/yt/yt.toml file.  This function is
        parallel-safe.

        Parameters
        ----------
        title : string
            The title for the model on the website
        description : string
            How you want the model to be described on the website
        api_key : string
            Optional; defaults to using the one in the config file
        color_field : string
            If specified, the field by which the surface will be colored
        color_map : string
            The name of the color map to use to map the color field
        color_log : bool
            Should the field be logged before being mapped to RGB?
        bounds : list of tuples
            [ (xmin, xmax), (ymin, ymax), (zmin, zmax) ] within which the model
            will be scaled and centered.  Defaults to the full domain.

        Returns
        -------
        URL : string
            The URL at which your model can be viewed.

        Examples
        --------

        >>> from yt.units import kpc
        >>> dd = ds.sphere("max", (200, "kpc"))
        >>> rho = 5e-27
        >>> bounds = [
        ...     (dd.center[i] - 100.0 * kpc, dd.center[i] + 100.0 * kpc)
        ...     for i in range(3)
        ... ]
        ...
        >>> surf = ds.surface(dd, ("gas", "density"), rho)
        >>> rv = surf.export_sketchfab(
        ...     title="Testing Upload",
        ...     description="A simple test of the uploader",
        ...     color_field="temperature",
        ...     color_map="hot",
        ...     color_log=True,
        ...     bounds=bounds,
        ... )
        ...
        """
        if color_map is None:
            color_map = ytcfg.get("yt", "default_colormap")
        api_key = api_key or ytcfg.get("yt", "sketchfab_api_key")
        if api_key in (None, "None"):
            raise YTNoAPIKey("SketchFab.com", "sketchfab_api_key")

        ply_file = TemporaryFile()
        self.export_ply(
            ply_file,
            bounds,
            color_field,
            color_map,
            color_log,
            sample_type="vertex",
            no_ghost=no_ghost,
        )
        ply_file.seek(0)
        # Greater than ten million vertices and we throw an error but dump
        # to a file.
        if self.vertices.shape[1] > 1e7:
            tfi = 0
            fn = "temp_model_%03i.ply" % tfi
            while os.path.exists(fn):
                fn = "temp_model_%03i.ply" % tfi
                tfi += 1
            open(fn, "wb").write(ply_file.read())
            raise YTTooManyVertices(self.vertices.shape[1], fn)

        zfs = NamedTemporaryFile(suffix=".zip")
        with zipfile.ZipFile(zfs, "w", zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("yt_export.ply", ply_file.read())
        zfs.seek(0)

        zfs.seek(0)
        data = {
            "token": api_key,
            "name": title,
            "description": description,
            "tags": "yt",
        }
        files = {"modelFile": zfs}
        upload_id = self._upload_to_sketchfab(data, files)
        upload_id = self.comm.mpi_bcast(upload_id, root=0)
        return upload_id

    @parallel_root_only
    def _upload_to_sketchfab(self, data, files):
        from yt.utilities.on_demand_imports import _requests as requests

        SKETCHFAB_DOMAIN = "sketchfab.com"
        SKETCHFAB_API_URL = f"https://api.{SKETCHFAB_DOMAIN}/v2/models"
        SKETCHFAB_MODEL_URL = f"https://{SKETCHFAB_DOMAIN}/models/"

        try:
            r = requests.post(SKETCHFAB_API_URL, data=data, files=files, verify=False)
        except requests.exceptions.RequestException:
            mylog.exception("An error has occurred")
            return

        result = r.json()

        if r.status_code != requests.codes.created:
            mylog.error("Upload to SketchFab failed with error: %s", result)
            return

        model_uid = result["uid"]
        model_url = SKETCHFAB_MODEL_URL + model_uid
        if model_uid:
            mylog.info("Model uploaded to: %s", model_url)
        else:
            mylog.error("Problem uploading.")

        return model_uid


class YTOctree(YTSelectionContainer3D):
    """A 3D region with all the data filled into an octree. This container
    will deposit particle fields onto octs using a kernel and SPH smoothing.
    The octree is built in a depth-first fashion. Depth-first search (DFS)
    means that tree starts refining at the root node (this is the largest node
    which contains every particles) and refines as far as possible along each
    branch before backtracking.

    Parameters
    ----------
    right_edge : array_like
        The right edge of the region to be extracted.  Specify units by supplying
        a YTArray, otherwise code length units are assumed.
    left_edge : array_like
        The left edge of the region to be extracted.  Specify units by supplying
        a YTArray, otherwise code length units are assumed.
    n_ref: int
        This is the maximum number of particles per leaf in the resulting
        octree.
    ptypes: list
        This is the type of particles to include when building the tree. This
        will default to all particles.

    Examples
    --------

    octree = ds.octree(n_ref=64)
    x_positions_of_cells = octree[('index', 'x')]
    y_positions_of_cells = octree[('index', 'y')]
    z_positions_of_cells = octree[('index', 'z')]
    density_of_gas_in_cells = octree[('gas', 'density')]
    """

    _spatial = True
    _type_name = "octree"

    _con_args = ("left_edge", "right_edge", "n_ref")
    _container_fields = (
        ("index", "dx"),
        ("index", "dy"),
        ("index", "dz"),
        ("index", "x"),
        ("index", "y"),
        ("index", "z"),
        ("index", "depth"),
        ("index", "refined"),
        ("index", "sizes"),
        ("index", "positions"),
    )

    def __init__(
        self,
        left_edge=None,
        right_edge=None,
        n_ref=32,
        ptypes=None,
        ds=None,
        field_parameters=None,
    ):
        if field_parameters is None:
            center = None
        else:
            center = field_parameters.get("center", None)
        YTSelectionContainer3D.__init__(self, center, ds, field_parameters)

        self.left_edge = self._sanitize_edge(left_edge, ds.domain_left_edge)
        self.right_edge = self._sanitize_edge(right_edge, ds.domain_right_edge)
        self.n_ref = n_ref
        self.ptypes = self._sanitize_ptypes(ptypes)

        self._setup_data_source()
        self.tree

    def _generate_tree(self):
        positions = []
        for ptype in self.ptypes:
            positions.append(
                self._data_source[(ptype, "particle_position")]
                .in_units("code_length")
                .d
            )

        positions = (
            np.concatenate(positions) if len(positions) > 0 else np.array(positions)
        )
        if not positions.size:
            mylog.info("No particles found!")
            self._octree = None
            return

        mylog.info("Allocating Octree for %s particles", positions.shape[0])
        self._octree = CyOctree(
            positions,
            left_edge=self.left_edge.to("code_length").d,
            right_edge=self.right_edge.to("code_length").d,
            n_ref=self.n_ref,
        )
        mylog.info("Allocated %s nodes in octree", self._octree.num_nodes)
        mylog.info("Octree bound %s particles", self._octree.bound_particles)

        # Now we store the index data about the octree in the python container
        ds = self.ds
        pos = ds.arr(self._octree.node_positions, "code_length")
        self[("index", "positions")] = pos
        self[("index", "x")] = pos[:, 0]
        self[("index", "y")] = pos[:, 1]
        self[("index", "z")] = pos[:, 2]
        self[("index", "refined")] = self._octree.node_refined

        sizes = ds.arr(self._octree.node_sizes, "code_length")
        self[("index", "sizes")] = sizes
        self[("index", "dx")] = sizes[:, 0]
        self[("index", "dy")] = sizes[:, 1]
        self[("index", "dz")] = sizes[:, 2]
        self[("index", "depth")] = self._octree.node_depth

    @property
    def tree(self):
        """
        The Cython+Python octree instance
        """
        if hasattr(self, "_octree"):
            return self._octree

        self._generate_tree()
        return self._octree

    @property
    def sph_smoothing_style(self):
        smoothing_style = getattr(self.ds, "sph_smoothing_style", "scatter")
        return smoothing_style

    @property
    def sph_normalize(self):
        normalize = getattr(self.ds, "use_sph_normalization", "normalize")
        return normalize

    @property
    def sph_num_neighbors(self):
        num_neighbors = getattr(self.ds, "num_neighbors", 32)
        return num_neighbors

    def _sanitize_ptypes(self, ptypes):
        if ptypes is None:
            return ["all"]

        if not isinstance(ptypes, list):
            ptypes = [ptypes]

        for ptype in ptypes:
            if ptype not in self.ds.particle_types:
                mess = f"{ptype} not found. Particle type must "
                mess += "be in the dataset!"
                raise TypeError(mess)

        return ptypes

    def _setup_data_source(self):
        mylog.info(
            (
                "Allocating octree with spatial range "
                "[%.4e, %.4e, %.4e] code_length to "
                "[%.4e, %.4e, %.4e] code_length"
            ),
            *self.left_edge.to("code_length").d,
            *self.right_edge.to("code_length").d,
        )
        self._data_source = self.ds.region(self.center, self.left_edge, self.right_edge)

    def _sanitize_edge(self, edge, default):
        if edge is None:
            return default.copy()
        if not is_sequence(edge):
            edge = [edge] * len(self.ds.domain_left_edge)
        if len(edge) != len(self.ds.domain_left_edge):
            raise RuntimeError(
                "Length of edges must match the dimensionality of the dataset"
            )
        if hasattr(edge, "units"):
            if edge.units.registry is self.ds.unit_registry:
                return edge
            edge_units = edge.units
        else:
            edge_units = "code_length"
        return self.ds.arr(edge, edge_units)

    def get_data(self, fields=None):
        if fields is None:
            return

        if hasattr(self.ds, "_sph_ptypes"):
            sph_ptypes = self.ds._sph_ptypes
        else:
            sph_ptypes = tuple(
                value
                for value in self.ds.particle_types_raw
                if value in ["PartType0", "Gas", "gas", "io"]
            )

        if len(sph_ptypes) == 0:
            raise RuntimeError
        smoothing_style = self.sph_smoothing_style
        normalize = self.sph_normalize

        if fields[0] in sph_ptypes:
            units = self.ds._get_field_info(fields).units
            if smoothing_style == "scatter":
                self._scatter_smooth(fields, units, normalize)
            else:
                self._gather_smooth(fields, units, normalize)
        elif fields[0] == "index":
            return self[fields]
        else:
            raise NotImplementedError

    def _gather_smooth(self, fields, units, normalize):
        buff = np.zeros(self.tree.num_nodes, dtype="float64")

        # Again, attempt to load num_neighbors from the octree
        num_neighbors = self.sph_num_neighbors

        # For the gather approach we load up all of the data, this like other
        # gather approaches is not memory conservative but with spatial chunking
        # this can be fixed
        fields_to_get = [
            "particle_position",
            "density",
            "particle_mass",
            "smoothing_length",
            fields[1],
        ]
        all_fields = all_data(self.ds, fields[0], fields_to_get, kdtree=True)

        interpolate_sph_positions_gather(
            buff,
            all_fields["particle_position"],
            self[("index", "positions")],
            all_fields["smoothing_length"],
            all_fields["particle_mass"],
            all_fields["density"],
            all_fields[fields[1]].in_units(units),
            self.ds.index.kdtree,
            use_normalization=normalize,
            num_neigh=num_neighbors,
        )

        self[fields] = self.ds.arr(buff[~self[("index", "refined")]], units)

    def _scatter_smooth(self, fields, units, normalize):
        buff = np.zeros(self.tree.num_nodes, dtype="float64")

        if normalize:
            buff_den = np.zeros(buff.shape[0], dtype="float64")
        else:
            buff_den = np.empty(0)

        ptype = fields[0]
        pbar = tqdm(desc=f"Interpolating (scatter) SPH field {fields[0]}")
        for chunk in self._data_source.chunks([fields], "io"):
            px = chunk[(ptype, "particle_position_x")].to("code_length").d
            py = chunk[(ptype, "particle_position_y")].to("code_length").d
            pz = chunk[(ptype, "particle_position_z")].to("code_length").d
            hsml = chunk[(ptype, "smoothing_length")].to("code_length").d
            pmass = chunk[(ptype, "particle_mass")].to("code_mass").d
            pdens = chunk[(ptype, "density")].to("code_mass/code_length**3").d
            field_quantity = chunk[fields].to(units).d

            if px.shape[0] > 0:
                self.tree.interpolate_sph_cells(
                    buff,
                    buff_den,
                    px,
                    py,
                    pz,
                    pmass,
                    pdens,
                    hsml,
                    field_quantity,
                    use_normalization=normalize,
                )

            pbar.update(1)
        pbar.close()

        if normalize:
            normalization_1d_utility(buff, buff_den)

        self[fields] = self.ds.arr(buff[~self[("index", "refined")]], units)



import os

from yt.frontends.gadget.data_structures import GadgetHDF5Dataset
from yt.utilities.on_demand_imports import _h5py as h5py

from .fields import GizmoFieldInfo


class GizmoDataset(GadgetHDF5Dataset):
    _field_info_class = GizmoFieldInfo

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        need_groups = ["Header"]
        veto_groups = ["FOF", "Group", "Subhalo"]
        valid = True
        valid_fname = filename
        # If passed arg is a directory, look for the .0 file in that dir
        if os.path.isdir(filename):
            valid_files = []
            for f in os.listdir(filename):
                fname = os.path.join(filename, f)
                fext = os.path.splitext(fname)[-1]
                if (
                    (".0" in f)
                    and (fext not in {".ewah", ".kdtree"})
                    and os.path.isfile(fname)
                ):
                    valid_files.append(fname)
            if len(valid_files) == 0:
                valid = False
            elif len(valid_files) > 1:
                valid = False
            else:
                valid_fname = valid_files[0]
        try:
            fh = h5py.File(valid_fname, mode="r")
            valid = all(ng in fh["/"] for ng in need_groups) and not any(
                vg in fh["/"] for vg in veto_groups
            )
            dmetal = "/PartType0/Metallicity"
            if dmetal not in fh or fh[dmetal].shape[1] < 11:
                valid = False
            fh.close()
        except Exception:
            valid = False
            pass
        return valid

    def _set_code_unit_attributes(self):
        super()._set_code_unit_attributes()



import glob

from yt.data_objects.time_series import DatasetSeries
from yt.funcs import only_on_root
from yt.loaders import load
from yt.utilities.exceptions import YTUnidentifiedDataType
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.parallel_tools.parallel_analysis_interface import parallel_objects


class ExodusIISimulation(DatasetSeries):
    r"""
    Initialize an ExodusII Simulation object.

    Upon creation, the input directory is searched for valid ExodusIIDatasets.
    The get_time_series can be used to generate a DatasetSeries object.

    simulation_directory : str
        The directory that contain the simulation data.

    Examples
    --------
    >>> import yt
    >>> sim = yt.load_simulation("demo_second", "ExodusII")
    >>> sim.get_time_series()
    >>> for ds in sim:
    ...     print(ds.current_time)

    """

    def __init__(self, simulation_directory, find_outputs=False):
        self.simulation_directory = simulation_directory
        fn_pattern = f"{self.simulation_directory}/*"
        potential_outputs = glob.glob(fn_pattern)
        self.all_outputs = self._check_for_outputs(potential_outputs)
        self.all_outputs.sort(key=lambda obj: obj["filename"])

    def __iter__(self):
        for o in self._pre_outputs:
            fn, step = o
            ds = load(fn, step=step)
            self._setup_function(ds)
            yield ds

    def __getitem__(self, key):
        if isinstance(key, slice):
            if isinstance(key.start, float):
                return self.get_range(key.start, key.stop)
            # This will return a sliced up object!
            return DatasetSeries(self._pre_outputs[key], self.parallel)
        o = self._pre_outputs[key]
        fn, step = o
        o = load(fn, step=step)
        self._setup_function(o)
        return o

    def get_time_series(self, parallel=False, setup_function=None):
        r"""
        Instantiate a DatasetSeries object for a set of outputs.

        If no additional keywords given, a DatasetSeries object will be
        created with all potential datasets created by the simulation.

        Fine-level filtering is currently not implemented.

        """

        all_outputs = self.all_outputs
        ds_list = []
        for output in all_outputs:
            num_steps = output["num_steps"]
            fn = output["filename"]
            for step in range(num_steps):
                ds_list.append((fn, step))
        super().__init__(ds_list, parallel=parallel, setup_function=setup_function)

    def _check_for_outputs(self, potential_outputs):
        r"""
        Check a list of files to see if they are valid datasets.
        """

        only_on_root(
            mylog.info, "Checking %d potential outputs.", len(potential_outputs)
        )

        my_outputs = {}
        for my_storage, output in parallel_objects(
            potential_outputs, storage=my_outputs
        ):
            try:
                ds = load(output)
            except (FileNotFoundError, YTUnidentifiedDataType):
                mylog.error("Failed to load %s", output)
                continue
            my_storage.result = {"filename": output, "num_steps": ds.num_steps}

        my_outputs = [
            my_output for my_output in my_outputs.values() if my_output is not None
        ]
        return my_outputs



from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer
from yt.fields.tensor_fields import setup_stress_energy_ideal

from .cfields import SRHDFields

b_units = "code_magnetic"
pre_units = "code_mass / (code_length*code_time**2)"
erg_units = "code_mass / (code_length*code_time**2)"
rho_units = "code_mass / code_length**3"
mom_units = "code_mass / (code_length**2*code_time)"
vel_units = "code_velocity"
pot_units = "code_length**2/code_time**2"

psi_units = "code_mass**(1/2) / code_length**(3/2)"


class GAMERFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        # hydro fields on disk (GAMER outputs conservative variables)
        ("Dens", (rho_units, [], None)),
        ("MomX", (mom_units, ["momentum_density_x"], None)),
        ("MomY", (mom_units, ["momentum_density_y"], None)),
        ("MomZ", (mom_units, ["momentum_density_z"], None)),
        ("Engy", (erg_units, ["total_energy_density"], None)),
        ("Pote", (pot_units, ["gravitational_potential"], None)),
        # MHD fields on disk (CC=cell-centered)
        ("CCMagX", (b_units, [], "B_x")),
        ("CCMagY", (b_units, [], "B_y")),
        ("CCMagZ", (b_units, [], "B_z")),
        # psiDM fields on disk
        ("Real", (psi_units, ["psidm_real_part"], None)),
        ("Imag", (psi_units, ["psidm_imaginary_part"], None)),
        # particle fields on disk (deposited onto grids)
        ("ParDens", (rho_units, ["particle_density_on_grid"], None)),
        ("TotalDens", (rho_units, ["total_density_on_grid"], None)),
    )

    known_particle_fields: KnownFieldsT = (
        ("ParMass", ("code_mass", ["particle_mass"], None)),
        ("ParPosX", ("code_length", ["particle_position_x"], None)),
        ("ParPosY", ("code_length", ["particle_position_y"], None)),
        ("ParPosZ", ("code_length", ["particle_position_z"], None)),
        ("ParVelX", ("code_velocity", ["particle_velocity_x"], None)),
        ("ParVelY", ("code_velocity", ["particle_velocity_y"], None)),
        ("ParVelZ", ("code_velocity", ["particle_velocity_z"], None)),
        ("ParCreTime", ("code_time", ["particle_creation_time"], None)),
    )

    def __init__(self, ds, field_list):
        super().__init__(ds, field_list)

    # add primitive and other derived variables
    def setup_fluid_fields(self):
        pc = self.ds.units.physical_constants
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        unit_system = self.ds.unit_system
        unit_system.registry = self.ds.unit_registry  # TODO: Why do I need this?!

        if self.ds.srhd:

            c2 = pc.clight * pc.clight
            c = pc.clight.in_units("code_length / code_time")
            if self.ds.eos == 4:
                fgen = SRHDFields(self.ds.eos, 0.0, c.d)
            else:
                fgen = SRHDFields(self.ds.eos, self.ds.gamma, c.d)

            def _sound_speed(field, data):
                out = fgen.sound_speed(data["gamer", "Temp"].d)
                return data.ds.arr(out, "code_velocity").to(unit_system["velocity"])

            def _gamma(field, data):
                out = fgen.gamma_field(data["gamer", "Temp"].d)
                return data.ds.arr(out, "dimensionless")

            # coordinate frame density
            self.alias(
                ("gas", "frame_density"),
                ("gamer", "Dens"),
                units=unit_system["density"],
            )

            self.add_field(
                ("gas", "gamma"), sampling_type="cell", function=_gamma, units=""
            )

            # 4-velocity spatial components
            def four_velocity_xyz(u):
                def _four_velocity(field, data):
                    out = fgen.four_velocity_xyz(
                        data["gamer", f"Mom{u.upper()}"].d,
                        data["gamer", "Dens"].d,
                        data["gamer", "Temp"].d,
                    )
                    return data.ds.arr(out, "code_velocity").to(unit_system["velocity"])

                return _four_velocity

            for u in "xyz":
                self.add_field(
                    ("gas", f"four_velocity_{u}"),
                    sampling_type="cell",
                    function=four_velocity_xyz(u),
                    units=unit_system["velocity"],
                )

            # lorentz factor
            def _lorentz_factor(field, data):
                out = fgen.lorentz_factor(
                    data["gamer", "Dens"].d,
                    data["gamer", "MomX"].d,
                    data["gamer", "MomY"].d,
                    data["gamer", "MomZ"].d,
                    data["gamer", "Temp"].d,
                )
                return data.ds.arr(out, "dimensionless")

            self.add_field(
                ("gas", "lorentz_factor"),
                sampling_type="cell",
                function=_lorentz_factor,
                units="",
            )

            # velocity
            def velocity_xyz(v):
                def _velocity(field, data):
                    out = fgen.velocity_xyz(
                        data["gamer", "Dens"].d,
                        data["gamer", "MomX"].d,
                        data["gamer", "MomY"].d,
                        data["gamer", "MomZ"].d,
                        data["gamer", "Temp"].d,
                        data["gamer", f"Mom{v.upper()}"].d,
                    )
                    return data.ds.arr(out, "code_velocity").to(unit_system["velocity"])

                return _velocity

            for v in "xyz":
                self.add_field(
                    ("gas", f"velocity_{v}"),
                    sampling_type="cell",
                    function=velocity_xyz(v),
                    units=unit_system["velocity"],
                )

            # density
            def _density(field, data):
                dens = fgen.density(
                    data["gamer", "Dens"].d,
                    data["gamer", "MomX"].d,
                    data["gamer", "MomY"].d,
                    data["gamer", "MomZ"].d,
                    data["gamer", "Temp"].d,
                )
                return data.ds.arr(dens, rho_units).to(unit_system["density"])

            self.add_field(
                ("gas", "density"),
                sampling_type="cell",
                function=_density,
                units=unit_system["density"],
            )

            # pressure
            def _pressure(field, data):
                out = fgen.pressure(
                    data["gamer", "Dens"].d,
                    data["gamer", "MomX"].d,
                    data["gamer", "MomY"].d,
                    data["gamer", "MomZ"].d,
                    data["gamer", "Temp"].d,
                )
                return data.ds.arr(out, pre_units).to(unit_system["pressure"])

            # thermal energy per mass (i.e., specific)
            def _specific_thermal_energy(field, data):
                out = fgen.specific_thermal_energy(
                    data["gamer", "Dens"].d,
                    data["gamer", "MomX"].d,
                    data["gamer", "MomY"].d,
                    data["gamer", "MomZ"].d,
                    data["gamer", "Temp"].d,
                )
                return data.ds.arr(out, "code_length**2 / code_time**2").to(
                    unit_system["specific_energy"]
                )

            # total energy per mass
            def _specific_total_energy(field, data):
                E = data["gamer", "Engy"] + data["gamer", "Dens"] * c2
                return E / data["gamer", "Dens"]

            def _kinetic_energy_density(field, data):
                out = fgen.kinetic_energy_density(
                    data["gamer", "Dens"].d,
                    data["gamer", "MomX"].d,
                    data["gamer", "MomY"].d,
                    data["gamer", "MomZ"].d,
                    data["gamer", "Temp"].d,
                )
                return data.ds.arr(out, erg_units).to(unit_system["pressure"])

            self.add_field(
                ("gas", "kinetic_energy_density"),
                sampling_type="cell",
                function=_kinetic_energy_density,
                units=unit_system["pressure"],
            )

            self.add_field(
                ("gas", "kinetic_energy_density"),
                sampling_type="cell",
                function=_kinetic_energy_density,
                units=unit_system["pressure"],
            )

            self.add_field(
                ("gas", "sound_speed"),
                sampling_type="cell",
                function=_sound_speed,
                units=unit_system["velocity"],
            )

            def _mach_number(field, data):
                out = fgen.mach_number(
                    data["gamer", "Dens"].d,
                    data["gamer", "MomX"].d,
                    data["gamer", "MomY"].d,
                    data["gamer", "MomZ"].d,
                    data["gamer", "Temp"].d,
                )
                return data.ds.arr(out, "dimensionless")

            self.add_field(
                ("gas", "mach_number"),
                sampling_type="cell",
                function=_mach_number,
                units="",
            )

            setup_stress_energy_ideal(self)

        else:  # not RHD

            # density
            self.alias(
                ("gas", "density"), ("gamer", "Dens"), units=unit_system["density"]
            )

            # velocity
            def velocity_xyz(v):
                def _velocity(field, data):
                    return data["gas", f"momentum_density_{v}"] / data["gas", "density"]

                return _velocity

            for v in "xyz":
                self.add_field(
                    ("gas", f"velocity_{v}"),
                    sampling_type="cell",
                    function=velocity_xyz(v),
                    units=unit_system["velocity"],
                )

            # ====================================================
            # note that yt internal fields assume
            #    [specific_thermal_energy]   = [energy per mass]
            #    [kinetic_energy_density]    = [energy per volume]
            #    [magnetic_energy_density]   = [energy per volume]
            # and we further adopt
            #    [specific_total_energy]     = [energy per mass]
            #    [total_energy_density]      = [energy per volume]
            # ====================================================

            # thermal energy per volume
            def et(data):
                ek = (
                    0.5
                    * (
                        data["gamer", "MomX"] ** 2
                        + data["gamer", "MomY"] ** 2
                        + data["gamer", "MomZ"] ** 2
                    )
                    / data["gamer", "Dens"]
                )
                Et = data["gamer", "Engy"] - ek
                if self.ds.mhd:
                    # magnetic_energy is a yt internal field
                    Et -= data["gas", "magnetic_energy_density"]
                return Et

            # thermal energy per mass (i.e., specific)
            def _specific_thermal_energy(field, data):
                return et(data) / data["gamer", "Dens"]

            # total energy per mass
            def _specific_total_energy(field, data):
                return data["gamer", "Engy"] / data["gamer", "Dens"]

            # pressure
            def _pressure(field, data):
                return et(data) * (data.ds.gamma - 1.0)

        self.add_field(
            ("gas", "specific_thermal_energy"),
            sampling_type="cell",
            function=_specific_thermal_energy,
            units=unit_system["specific_energy"],
        )

        def _thermal_energy_density(field, data):
            return data["gas", "density"] * data["gas", "specific_thermal_energy"]

        self.add_field(
            ("gas", "thermal_energy_density"),
            sampling_type="cell",
            function=_thermal_energy_density,
            units=unit_system["pressure"],
        )

        self.add_field(
            ("gas", "specific_total_energy"),
            sampling_type="cell",
            function=_specific_total_energy,
            units=unit_system["specific_energy"],
        )

        self.add_field(
            ("gas", "pressure"),
            sampling_type="cell",
            function=_pressure,
            units=unit_system["pressure"],
        )

        # mean molecular weight
        if hasattr(self.ds, "mu"):

            def _mu(field, data):
                return data.ds.mu * data["index", "ones"]

            self.add_field(
                ("gas", "mean_molecular_weight"),
                sampling_type="cell",
                function=_mu,
                units="",
            )

        # temperature
        def _temperature(field, data):
            return (
                data.ds.mu
                * data["gas", "pressure"]
                * pc.mh
                / (data["gas", "density"] * pc.kb)
            )

        self.add_field(
            ("gas", "temperature"),
            sampling_type="cell",
            function=_temperature,
            units=unit_system["temperature"],
        )

        # magnetic field aliases --> magnetic_field_x/y/z
        if self.ds.mhd:
            setup_magnetic_field_aliases(self, "gamer", [f"CCMag{v}" for v in "XYZ"])

    def setup_particle_fields(self, ptype):
        super().setup_particle_fields(ptype)



import contextlib
import glob
import logging
import os
import platform
import shutil
import subprocess
import sys
import tempfile
from textwrap import dedent
from concurrent.futures import ThreadPoolExecutor
from distutils.ccompiler import CCompiler, new_compiler
from distutils.sysconfig import customize_compiler
from subprocess import PIPE, Popen
from sys import platform as _platform

from pkg_resources import resource_filename
from setuptools.command.build_ext import build_ext as _build_ext
from setuptools.command.sdist import sdist as _sdist
from setuptools.errors import CompileError, LinkError

log = logging.getLogger("setupext")

@contextlib.contextmanager
def stdchannel_redirected(stdchannel, dest_filename):
    """
    A context manager to temporarily redirect stdout or stderr

    e.g.:

    with stdchannel_redirected(sys.stderr, os.devnull):
        if compiler.has_function('clock_gettime', libraries=['rt']):
            libraries.append('rt')

    Code adapted from https://stackoverflow.com/a/17752455/1382869
    """

    try:
        oldstdchannel = os.dup(stdchannel.fileno())
        dest_file = open(dest_filename, "w")
        os.dup2(dest_file.fileno(), stdchannel.fileno())

        yield
    finally:
        if oldstdchannel is not None:
            os.dup2(oldstdchannel, stdchannel.fileno())
        if dest_file is not None:
            dest_file.close()


def check_for_openmp():
    """Returns OpenMP compiler and linker flags if local setup supports
    OpenMP or [], [] otherwise

    Code adapted from astropy_helpers, originally written by Tom
    Robitaille and Curtis McCully.
    """

    # Create a temporary directory
    ccompiler = new_compiler()
    customize_compiler(ccompiler)

    tmp_dir = tempfile.mkdtemp()
    start_dir = os.path.abspath(".")

    CCODE = dedent("""\
        #include <omp.h>
        #include <stdio.h>
        int main() {
            omp_set_num_threads(2);
            #pragma omp parallel
            printf("nthreads=%d\\n", omp_get_num_threads());
            return 0;
        }"""
    )

    # TODO: test more known compilers:
    # MinGW, AppleClang with libomp, MSVC, ICC, XL, PGI, ...
    if os.name == "nt":
        # TODO: make this work with mingw
        # AFAICS there's no easy way to get the compiler distutils
        # will be using until compilation actually happens
        compile_flags = ["-openmp"]
        link_flags = [""]
    else:
        compile_flags = ["-fopenmp"]
        link_flags = ["-fopenmp"]

    try:
        os.chdir(tmp_dir)

        with open("test_openmp.c", "w") as f:
            f.write(CCODE)

        os.mkdir("objects")

        # Compile, link, and run test program
        with stdchannel_redirected(sys.stderr, os.devnull):
            ccompiler.compile(
                ["test_openmp.c"], output_dir="objects", extra_postargs=compile_flags
            )
            ccompiler.link_executable(
                glob.glob(os.path.join("objects", "*")),
                "test_openmp",
                extra_postargs=link_flags,
            )
            output = (
                subprocess.check_output("./test_openmp")
                .decode(sys.stdout.encoding or "utf-8")
                .splitlines()
            )

        if "nthreads=" in output[0]:
            nthreads = int(output[0].strip().split("=")[1])
            if len(output) == nthreads:
                using_openmp = True
            else:
                log.warning(
                    "Unexpected number of lines from output of test "
                    "OpenMP program (output was %s)",
                    output,
                )
                using_openmp = False
        else:
            log.warning(
                "Unexpected output from test OpenMP program (output was %s)", output
            )
            using_openmp = False

    except (CompileError, LinkError):
        using_openmp = False
    finally:
        os.chdir(start_dir)

    if using_openmp:
        log.warning("Using OpenMP to compile parallel extensions")
    else:
        log.warning(
            "Unable to compile OpenMP test program so Cython\n"
            "extensions will be compiled without parallel support"
        )

    if using_openmp:
        return compile_flags, link_flags
    else:
        return [], []


def check_CPP14_flag(compile_flags):
    # Create a temporary directory
    ccompiler = new_compiler()
    customize_compiler(ccompiler)

    tmp_dir = tempfile.mkdtemp()
    start_dir = os.path.abspath(".")

    # Note: This code requires C++14 functionalities (also required to compile yt)
    # It compiles on gcc 4.7.4 (together with the entirety of yt) with the flag "-std=gnu++0x".
    # It does not compile on gcc 4.6.4 (neither does yt).
    CPPCODE = dedent("""\
        #include <vector>

        struct node {
            std::vector<int> vic;
            bool visited = false;
        };

        int main() {
            return 0;
        }"""
    )

    os.chdir(tmp_dir)
    try:
        with open("test_cpp14.cpp", "w") as f:
            f.write(CPPCODE)

        os.mkdir("objects")

        # Compile, link, and run test program
        with stdchannel_redirected(sys.stderr, os.devnull):
            ccompiler.compile(
                ["test_cpp14.cpp"], output_dir="objects", extra_postargs=compile_flags
            )
        return True
    except CompileError:
        return False
    finally:
        os.chdir(start_dir)


def check_CPP14_flags(possible_compile_flags):
    for flags in possible_compile_flags:
        if check_CPP14_flag([flags]):
            return flags

    log.warning(
        "Your compiler seems to be too old to support C++14. "
        "yt may not be able to compile. Please use a newer version."
    )
    return []


def check_for_pyembree(std_libs):
    embree_libs = []
    embree_aliases = {}
    try:
        _ = resource_filename("pyembree", "rtcore.pxd")
    except ImportError:
        return embree_libs, embree_aliases

    embree_prefix = os.path.abspath(read_embree_location())
    embree_inc_dir = os.path.join(embree_prefix, "include")
    embree_lib_dir = os.path.join(embree_prefix, "lib")

    if _platform == "darwin":
        embree_lib_name = "embree.2"
    else:
        embree_lib_name = "embree"

    embree_aliases["EMBREE_INC_DIR"] = ["yt/utilities/lib/", embree_inc_dir]
    embree_aliases["EMBREE_LIB_DIR"] = [embree_lib_dir]
    embree_aliases["EMBREE_LIBS"] = std_libs + [embree_lib_name]
    embree_libs += ["yt/utilities/lib/embree_mesh/*.pyx"]

    if in_conda_env():
        conda_basedir = os.path.dirname(os.path.dirname(sys.executable))
        embree_aliases["EMBREE_INC_DIR"].append(os.path.join(conda_basedir, "include"))
        embree_aliases["EMBREE_LIB_DIR"].append(os.path.join(conda_basedir, "lib"))

    return embree_libs, embree_aliases


def in_conda_env():
    return any(s in sys.version for s in ("Anaconda", "Continuum", "conda-forge"))


def read_embree_location():
    """

    Attempts to locate the embree installation. First, we check for an
    EMBREE_DIR environment variable. If one is not defined, we look for
    an embree.cfg file in the root yt source directory. Finally, if that
    is not present, we default to /usr/local. If embree is installed in a
    non-standard location and none of the above are set, the compile will
    not succeed. This only gets called if check_for_pyembree() returns
    something other than None.

    """

    rd = os.environ.get("EMBREE_DIR")
    if rd is None:
        try:
            rd = open("embree.cfg").read().strip()
        except IOError:
            rd = "/usr/local"

    fail_msg = (
        "I attempted to find Embree headers in %s. \n"
        "If this is not correct, please set your correct embree location \n"
        "using EMBREE_DIR environment variable or your embree.cfg file. \n"
        "Please see http://yt-project.org/docs/dev/visualizing/unstructured_mesh_rendering.html "
        "for more information. \n" % rd
    )

    # Create a temporary directory
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()

    try:
        os.chdir(tmpdir)

        # Get compiler invocation
        compiler = os.getenv("CXX", "c++")
        compiler = compiler.split(" ")

        # Attempt to compile a test script.
        filename = r"test.cpp"
        file = open(filename, "wt", 1)
        CCODE = dedent("""\
            #include "embree2/rtcore.h
            int main() {
                return 0;
            }"""
        )
        file.write(CCODE)
        file.flush()
        p = Popen(
            compiler + ["-I%s/include/" % rd, filename],
            stdin=PIPE,
            stdout=PIPE,
            stderr=PIPE,
        )
        output, err = p.communicate()
        exit_code = p.returncode

        if exit_code != 0:
            log.warning(
                "Pyembree is installed, but I could not compile Embree test code."
            )
            log.warning("The error message was: ")
            log.warning(err)
            log.warning(fail_msg)

        # Clean up
        file.close()

    except OSError:
        log.warning(
            "read_embree_location() could not find your C compiler. "
            "Attempted to use '%s'.",
            compiler,
        )
        return False

    finally:
        os.chdir(curdir)
        shutil.rmtree(tmpdir)

    return rd


def get_cpu_count():
    if platform.system() == "Windows":
        return 0

    cpu_count = os.cpu_count()
    try:
        user_max_cores = int(os.getenv("MAX_BUILD_CORES", cpu_count))
    except ValueError as e:
        raise ValueError(
            "MAX_BUILD_CORES must be set to an integer. "
            + "See above for original error."
        ) from e
    max_cores = min(cpu_count, user_max_cores)
    return max_cores


def install_ccompiler():
    def _compile(
        self,
        sources,
        output_dir=None,
        macros=None,
        include_dirs=None,
        debug=0,
        extra_preargs=None,
        extra_postargs=None,
        depends=None,
    ):
        """Function to monkey-patch distutils.ccompiler.CCompiler"""
        macros, objects, extra_postargs, pp_opts, build = self._setup_compile(
            output_dir, macros, include_dirs, sources, depends, extra_postargs
        )
        cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)

        for obj in objects:
            try:
                src, ext = build[obj]
            except KeyError:
                continue
            self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)

        # Return *all* object filenames, not just the ones we just built.
        return objects

    CCompiler.compile = _compile


def create_build_ext(lib_exts, cythonize_aliases):
    class build_ext(_build_ext):
        # subclass setuptools extension builder to avoid importing cython and numpy
        # at top level in setup.py. See http://stackoverflow.com/a/21621689/1382869
        # NOTE: this is likely not necessary anymore since
        # pyproject.toml was introduced in the project

        def finalize_options(self):
            from Cython.Build import cythonize

            # Override the list of extension modules
            self.distribution.ext_modules[:] = cythonize(
                lib_exts,
                aliases=cythonize_aliases,
                compiler_directives={"language_level": 2},
                nthreads=get_cpu_count(),
            )
            _build_ext.finalize_options(self)
            # Prevent numpy from thinking it is still in its setup process
            # see http://stackoverflow.com/a/21621493/1382869
            if isinstance(__builtins__, dict):
                # sometimes this is a dict so we need to check for that
                # https://docs.python.org/3/library/builtins.html
                __builtins__["__NUMPY_SETUP__"] = False
            else:
                __builtins__.__NUMPY_SETUP__ = False
            import numpy

            self.include_dirs.append(numpy.get_include())

        def build_extensions(self):
            self.check_extensions_list(self.extensions)

            ncpus = get_cpu_count()
            if ncpus > 0:
                with ThreadPoolExecutor(ncpus) as executor:
                    results = {
                        executor.submit(self.build_extension, extension): extension
                        for extension in self.extensions
                    }
                for result in results:
                    result.result()
            else:
                super().build_extensions()

        def build_extension(self, extension):
            try:
                super().build_extension(extension)
            except CompileError as exc:
                print(f"While building '{extension.name}' following error was raised:\n {exc}")
                raise

    class sdist(_sdist):
        # subclass setuptools source distribution builder to ensure cython
        # generated C files are included in source distribution.
        # See http://stackoverflow.com/a/18418524/1382869
        def run(self):
            # Make sure the compiled Cython files in the distribution are up-to-date
            from Cython.Build import cythonize

            cythonize(
                lib_exts,
                aliases=cythonize_aliases,
                compiler_directives={"language_level": 2},
                nthreads=get_cpu_count(),
            )
            _sdist.run(self)

    return build_ext, sdist



from yt.frontends.gamer.api import GAMERDataset
from yt.testing import (
    assert_array_almost_equal,
    assert_equal,
    requires_file,
    units_override_check,
)
from yt.utilities.answer_testing.framework import (
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

jet = "InteractingJets/jet_000002"
_fields_jet = (
    ("gas", "temperature"),
    ("gas", "density"),
    ("gas", "velocity_magnitude"),
)
jet_units = {
    "length_unit": (1.0, "kpc"),
    "time_unit": (3.08567758096e13, "s"),
    "mass_unit": (1.4690033e36, "g"),
}


@requires_ds(jet, big_data=True)
def test_jet():
    ds = data_dir_load(jet, kwargs={"units_override": jet_units})
    assert_equal(str(ds), "jet_000002")
    for test in small_patch_amr(ds, _fields_jet):
        test_jet.__name__ = test.description
        yield test


psiDM = "WaveDarkMatter/psiDM_000020"
_fields_psiDM = ("Dens", "Real", "Imag")


@requires_ds(psiDM, big_data=True)
def test_psiDM():
    ds = data_dir_load(psiDM)
    assert_equal(str(ds), "psiDM_000020")
    for test in small_patch_amr(ds, _fields_psiDM):
        test_psiDM.__name__ = test.description
        yield test


plummer = "Plummer/plummer_000000"
_fields_plummer = (("gamer", "ParDens"), ("deposit", "io_cic"))


@requires_ds(plummer, big_data=True)
def test_plummer():
    ds = data_dir_load(plummer)
    assert_equal(str(ds), "plummer_000000")
    for test in small_patch_amr(ds, _fields_plummer):
        test_plummer.__name__ = test.description
        yield test


mhdvortex = "MHDOrszagTangVortex/Data_000018"
_fields_mhdvortex = (
    ("gamer", "CCMagX"),
    ("gamer", "CCMagY"),
    ("gas", "magnetic_energy_density"),
)


@requires_ds(mhdvortex, big_data=True)
def test_mhdvortex():
    ds = data_dir_load(mhdvortex)
    assert_equal(str(ds), "Data_000018")
    for test in small_patch_amr(ds, _fields_mhdvortex):
        test_mhdvortex.__name__ = test.description
        yield test


@requires_file(psiDM)
def test_GAMERDataset():
    assert isinstance(data_dir_load(psiDM), GAMERDataset)


@requires_file(jet)
def test_units_override():
    units_override_check(jet)


jiw = "JetICMWall/Data_000060"
_fields_jiw = (
    ("gas", "four_velocity_magnitude"),
    ("gas", "density"),
    ("gas", "gamma"),
    ("gas", "temperature"),
)


@requires_ds(jiw, big_data=True)
def test_jiw():
    ds = data_dir_load(jiw)
    assert_equal(str(ds), "Data_000060")
    for test in small_patch_amr(ds, _fields_jiw):
        test_jiw.__name__ = test.description
        yield test


@requires_ds(jiw, big_data=True)
def test_stress_energy():
    axes = "txyz"
    ds = data_dir_load(jiw)
    center = ds.arr([3.0, 10.0, 10.0], "kpc")
    sp = ds.sphere(center, (1.0, "kpc"))
    c2 = ds.units.clight**2
    inv_c2 = 1.0 / c2
    rho = sp["gas", "density"]
    p = sp["gas", "pressure"]
    e = sp["gas", "thermal_energy_density"]
    h = rho + (e + p) * inv_c2
    for mu in range(4):
        for nu in range(4):
            # matrix is symmetric so only do the upper-right part
            if nu >= mu:
                Umu = sp[f"four_velocity_{axes[mu]}"]
                Unu = sp[f"four_velocity_{axes[nu]}"]
                Tmunu = h * Umu * Unu
                if mu != nu:
                    assert_array_almost_equal(sp[f"T{mu}{nu}"], sp[f"T{nu}{mu}"])
                else:
                    Tmunu += p
                assert_array_almost_equal(sp[f"T{mu}{nu}"], Tmunu)



from . import tests
from .data_structures import GadgetDataset, GadgetFieldInfo, GadgetHDF5Dataset
from .io import IOHandlerGadgetBinary, IOHandlerGadgetHDF5
from .simulation_handling import GadgetSimulation



from nose.plugins.attrib import attr

from yt.testing import ANSWER_TEST_TAG, fake_particle_ds
from yt.utilities.answer_testing.framework import GenericImageTest
from yt.visualization.api import ParticleProjectionPlot


@attr(ANSWER_TEST_TAG)
def test_particleprojectionplot_set_colorbar_properties():
    ds = fake_particle_ds(npart=100)

    def create_image(filename_prefix):
        field = ("all", "particle_mass")
        p = ParticleProjectionPlot(ds, 2, field)
        p.set_buff_size(10)

        p.set_unit(field, "Msun")
        p.save(f"{filename_prefix}_set_unit")

        p.set_zlim(field, zmax=1e-35)
        p.save(f"{filename_prefix}_set_unit_zlim")

        p.set_log(field, False)
        p.save(f"{filename_prefix}_set_unit_zlim_log")

    test = GenericImageTest(ds, create_image, 12)
    test.prefix = "test_particleprojectionplot_set_colorbar_properties"
    test.answer_name = "particleprojectionplot_set_colorbar_properties"
    yield test



import numpy as np

from yt.testing import assert_allclose_units, fake_amr_ds


def test_AM_value():
    ds = fake_amr_ds(
        fields=("Density", "velocity_x", "velocity_y", "velocity_z"),
        units=("g/cm**3", "cm/s", "cm/s", "cm/s"),
        length_unit=0.5,
    )

    sp = ds.sphere([0.5] * 3, (0.1, "code_length"))

    x0 = sp.center
    v0 = ds.arr([1, 2, 3], "km/s")

    sp.set_field_parameter("bulk_velocity", v0)

    X = (ds.arr([sp[("index", k)] for k in "xyz"]) - x0[:, None]).T
    V = (ds.arr([sp[("gas", f"velocity_{k}")] for k in "xyz"]) - v0[:, None]).T

    sAM_manual = ds.arr(np.cross(X, V), X.units * V.units)
    sAM = ds.arr([sp[("gas", f"specific_angular_momentum_{k}")] for k in "xyz"]).T

    assert_allclose_units(sAM_manual, sAM)



from yt.fields.field_info_container import FieldInfoContainer

# The nice thing about GDF is that for the most part, everything is in CGS,
# with potentially a scalar modification.


class GDFFieldInfo(FieldInfoContainer):
    known_other_fields = (
        ("density", ("g/cm**3", ["density"], None)),
        ("specific_energy", ("erg/g", ["specific_thermal_energy"], None)),
        ("pressure", ("erg/cm**3", ["pressure"], None)),
        ("temperature", ("K", ["temperature"], None)),
        ("velocity_x", ("cm/s", ["velocity_x"], None)),
        ("velocity_y", ("cm/s", ["velocity_y"], None)),
        ("velocity_z", ("cm/s", ["velocity_z"], None)),
        ("mag_field_x", ("gauss", ["magnetic_field_x"], None)),
        ("mag_field_y", ("gauss", ["magnetic_field_y"], None)),
        ("mag_field_z", ("gauss", ["magnetic_field_z"], None)),
    )
    known_particle_fields = ()

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        setup_magnetic_field_aliases(
            self, "gdf", [f"magnetic_field_{ax}" for ax in "xyz"]
        )



import yt

# Load the dataset.
ds = yt.load("GalaxyClusterMerger/fiducial_1to3_b0.273d_hdf5_plt_cnt_0175")

# Create density-weighted projections of standard deviation of the velocity
# (weighted line integrals)

for normal in "xyz":
    yt.ProjectionPlot(
        ds,
        normal,
        ("gas", f"velocity_{normal}"),
        weight_field=("gas", "density"),
        moment=2,
    ).save()
